                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 9: garbage collection   */
                     /*-----------------------------------*/
                     /*              compiler             */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipGrammar.h"

#include "SlipDictionary.h"
#include "SlipPool.h"
#include "SlipRead.h"
#include "SlipStack.h"

#include "SlipCompile.h"

/*----------------------------- private constants ----------------------------*/

static const TXT_type AL1_message = "at least one expression required";
static const TXT_type E1X_message = "exactly one expression required";
static const TXT_type ITF_message = "improperly terminated form";
static const TXT_type IPA_message = "invalid parameter";
static const TXT_type IVP_message = "invalid pattern in form";
static const TXT_type IVV_message = "missing variable in form";
static const TXT_type IXT_message = "invalid expression type";
static const TXT_type MSP_message = "missing pattern in form";
static const TXT_type MSV_message = "invalid variable in form";
static const TXT_type PNR_message = "path name required";
static const TXT_type TFX_message = "too few expressions in form";
static const TXT_type TMX_message = "too many expressions in form";

/*---------------------------- private prototypes ----------------------------*/

static EXP_type compile_expression_C(EXP_type);
static EXP_type compile_parameters_C(PAI_type,
                                     TXT_type);
static VEC_type   compile_sequence_C(PAI_type,
                                     TXT_type);

/*----------------------------- private functions ----------------------------*/

static EXP_type error(TXT_type Message)
  { Main_Error(Message);
    return Grammar_Unspecified; }

static EXP_type error_tag(TXT_type Message,
                          TAG_type Tag)
  { sprintf(Main_Buffer, "%s: %d", Message, Tag);
    return error(Main_Buffer); }

static EXP_type error_text(TXT_type Message,
                           TXT_type Text)
  { sprintf(Main_Buffer, "%s: %s", Message, Text);
    return error(Main_Buffer); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- application -------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_application_C(PAI_type Application)
  { APL_type compiled_application;
    EXP_type compiled_operator,
             operands,
             operator;
    PAI_type application;
    TAG_type tag;
    VEC_type compiled_operands;
    operands = Application->cdr;
    tag = Grammar_Tag(operands);
    switch (tag)
      { case NUL_tag:
          operator = Application->car;
          compiled_operator = compile_expression_C(operator);
          SURVIVE_DEFAULT_C(compiled_operator);
          compiled_operands = Grammar_Empty_Vector_P;
          break;
        case PAI_tag:
          application = Stack_Push_C(Application);
          operator = application->car;
          compiled_operator = compile_expression_C(operator);
          application = Stack_Peek();
          operands = application->cdr;
          Stack_Poke(compiled_operator);
          compiled_operands = compile_sequence_C(operands,
                                                 Pool_Application_Text);
          SURVIVE_DEFAULT_C(compiled_operands);
          compiled_operator = Stack_Pop();
          break;
        default:
          return error(ITF_message); }
    compiled_application = make_APL_M(compiled_operator,
                                      compiled_operands);
    return compiled_application; }

/*----------------------------------------------------------------------------*/
/*----------------------------------- begin ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_begin_C(PAI_type Operands)
  { BEG_type compiled_begin;
    TAG_type tag;
    TXT_type message;
    VEC_type compiled_sequence;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          message = AL1_message;
          goto error;
        case PAI_tag:
          compiled_sequence = compile_sequence_C(Operands,
                                                 Pool_Begin_Text);
          SURVIVE_DEFAULT_C(compiled_sequence);
          compiled_begin = make_BEG_M(compiled_sequence);
          return compiled_begin;
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_Begin_Text); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- define ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_define_function_C(EXP_type Operands)
  { EXP_type residue;
    DFF_type compiled_define_function;
    DFZ_type compiled_define_function_vararg;
    NBR_type frame_size,
             offset,
             param_count;
    PAI_type body,
             operands,
             parameters,
             pattern;
    SYM_type symbol,
             vararg;
    TAG_type tag;
    TXT_type message;
    UNS_type raw_frame_size,
             raw_offset,
             raw_param_count;
    VEC_type compiled_body;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          message = MSP_message;
          goto error;
        case PAI_tag:
          operands = Stack_Push_C(Operands);
          pattern = operands->car;
          symbol = pattern->car;
          if (!is_SYM(symbol))
            { message = IVV_message;
              goto error; }
          SURVIVE_DEFAULT_C(symbol);
          raw_offset = Dictionary_Define_Symbol_M(symbol,
                                                  TRUE);
          Dictionary_Enter_Nested_Scope_M();
          operands = Stack_Peek();
          pattern = operands->car;
          parameters = pattern->cdr;
          residue = compile_parameters_C(parameters,
                                         Pool_Define_Text);
          operands = Stack_Peek();
          pattern = operands->car;
          body    = operands->cdr;
          if (!is_PAI(body))
            { message = AL1_message;
              goto error; }
          symbol = pattern->car;
          Stack_Poke(symbol);
          raw_param_count = Dictionary_Get_Frame_Size();
          tag = Grammar_Tag(residue);
          switch (tag)
            { case NUL_tag:
                compiled_body = compile_sequence_C(body,
                                                   Pool_Define_Text);
                raw_frame_size = Dictionary_Exit_Nested_Scope(); // PROPAGATE!!!
                SURVIVE_DEFAULT_C(compiled_body);
                offset = make_NBR(raw_offset);
                param_count = make_NBR(raw_param_count);
                frame_size = make_NBR(raw_frame_size);
                symbol = Stack_Pop();
                compiled_define_function = make_DFF_M(symbol,
                                                      offset,
                                                      param_count,
                                                      frame_size,
                                                      compiled_body);
                return compiled_define_function;
              case SYM_tag:
                Stack_Push_C(body);
                SURVIVE_DEFAULT_C(residue);
                vararg = residue;
                Dictionary_Define_Symbol_M(vararg,
                                           FALSE);
                body = Stack_Pop();
                compiled_body = compile_sequence_C(body,
                                                   Pool_Define_Text);
                raw_frame_size = Dictionary_Exit_Nested_Scope();
                SURVIVE_DEFAULT_C(compiled_body);
                offset = make_NBR(raw_offset);
                param_count = make_NBR(raw_param_count);
                frame_size = make_NBR(raw_frame_size);
                symbol = Stack_Pop();
                compiled_define_function_vararg = make_DFZ_M(symbol,
                                                             offset,
                                                             param_count,
                                                             frame_size,
                                                             compiled_body);
                return compiled_define_function_vararg;
            default:
              message = IPA_message;
              goto error; }
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_Define_Text); }

static DFV_type compile_define_variable_C(PAI_type Operands)
  { DFV_type compiled_definition;
    EXP_type compiled_expression,
             expression,
             residue;
    NBR_type offset;
    PAI_type expressions,
             operands;
    SYM_type symbol;
    TAG_type tag;
    TXT_type message;
    UNS_type raw_offset;
    expressions = Operands->cdr;
    tag = Grammar_Tag(expressions);
    switch (tag)
      { case NUL_tag:
          message = TFX_message;
          goto error;
        case PAI_tag:
          residue = expressions->cdr;
          tag = Grammar_Tag(residue);
          switch (tag)
            { case NUL_tag:
                operands = Stack_Push_C(Operands);
                expressions = operands->cdr;
                expression = expressions->car;
                compiled_expression = compile_expression_C(expression);
                SURVIVE_DEFAULT_C(compiled_expression);
                operands = Stack_Pop();
                symbol = operands->car;
                raw_offset = Dictionary_Define_Symbol_M(symbol,
                                                        TRUE);
                offset = make_NBR(raw_offset);
                compiled_definition = make_DFV_M(offset,
                                                 compiled_expression);
                return compiled_definition;
              case PAI_tag:
                message = TMX_message;
                goto error;
              default:
                message = ITF_message;
                goto error; }
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_Define_Text); }

static EXP_type compile_define_C(PAI_type Operands)
  { EXP_type compiled_definition,
             pattern;
    TAG_type tag;
    TXT_type message;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          message = MSP_message;
          goto error;
        case PAI_tag:
          pattern = Operands->car;
          tag = Grammar_Tag(pattern);
          switch (tag)
            { case PAI_tag:
                compiled_definition = compile_define_function_C(Operands);
                return compiled_definition;
              case SYM_tag:
                compiled_definition = compile_define_variable_C(Operands);
                return compiled_definition;
              default:
                message = IVP_message;
                goto error; }
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_Define_Text); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ if ------------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_if_C(PAI_type Operands)
  { EXP_type alternative,
             compiled_alternative,
             compiled_consequent,
             compiled_predicate,
             consequent,
             predicate,
             residue;
    IFF_type compiled_double_if;
    IFZ_type compiled_single_if;
    NBR_type alternative_size,
             consequent_size;
    PAI_type alternatives,
             clauses,
             operands;
    TAG_type tag;
    TXT_type message;
    UNS_type raw_alternative_size,
             raw_consequent_size;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          message = TFX_message;
          goto error;
        case PAI_tag:
          clauses = Operands->cdr;
          tag = Grammar_Tag(clauses);
          switch (tag)
            { case NUL_tag:
                message = TFX_message;
                goto error;
              case PAI_tag:
                operands = Stack_Push_C(Operands);
                predicate = operands->car;
                compiled_predicate = compile_expression_C(predicate);
                SURVIVE_DEFAULT_C(compiled_predicate);
                Dictionary_Enter_Nested_Scope_M();
                operands = Stack_Peek();
                Stack_Poke(compiled_predicate);
                clauses = operands->cdr;
                alternatives = clauses->cdr;
                tag = Grammar_Tag(alternatives);
                switch (tag)
                  { case NUL_tag:
                      consequent = clauses->car;
                      compiled_consequent = compile_expression_C(consequent);
                      raw_consequent_size = Dictionary_Exit_Nested_Scope();
                      SURVIVE_DEFAULT_C(compiled_consequent);
                      compiled_predicate = Stack_Pop();
                      consequent_size = make_NBR(raw_consequent_size);
                      compiled_single_if = make_IFZ_M(compiled_predicate,
                                                      compiled_consequent,
                                                      consequent_size);
                      return compiled_single_if;
                    case PAI_tag:
                      residue = alternatives->cdr;
                      if (is_NUL(residue))
                        { clauses = Stack_Push_C(clauses);
                          consequent = clauses->car;
                          compiled_consequent = compile_expression_C(consequent);
                          raw_consequent_size = Dictionary_Exit_Nested_Scope();
                          SURVIVE_DEFAULT_C(compiled_consequent);
                          Dictionary_Enter_Nested_Scope_M();
                          clauses = Stack_Peek();
                          alternatives = clauses->cdr;
                          alternative = alternatives->car;
                          Stack_Poke(compiled_consequent);
                          compiled_alternative = compile_expression_C(alternative);
                          raw_alternative_size = Dictionary_Exit_Nested_Scope();
                          SURVIVE_DEFAULT_C(compiled_alternative);
                          compiled_consequent = Stack_Pop();
                          compiled_predicate = Stack_Pop();
                          consequent_size = make_NBR(raw_consequent_size);
                          alternative_size = make_NBR(raw_alternative_size);
                          compiled_double_if = make_IFF_M(compiled_predicate,
                                                          compiled_consequent,
                                                          consequent_size,
                                                          compiled_alternative,
                                                          alternative_size);
                          return compiled_double_if; }
                        else
                          { message = TMX_message;
                            goto error; }
                   default:
                     message = ITF_message;
                     goto error; }
              default:
                message = ITF_message;
                goto error; }
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_If_Text); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ lambda --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_lambda_C(PAI_type Operands)
  { EXP_type parameters,
             residue;
    LMB_type compiled_lambda;
    LMZ_type compiled_lambda_vararg;
    NBR_type frame_size,
             param_count;
    PAI_type body,
             operands;
    TAG_type tag;
    TXT_type message;
    UNS_type raw_frame_size,
             raw_param_count;
    VEC_type compiled_body;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          message = TFX_message;
          goto error;
        case PAI_tag:
          Stack_Push_C(Operands);
          Grammar_Default_Claim_C();
          Dictionary_Enter_Nested_Scope_M();
          operands = Stack_Peek();
          parameters = operands->car;
          body       = operands->cdr;
          Stack_Poke(body);
          residue = compile_parameters_C(parameters,
                                         Pool_Lambda_Text);
          raw_param_count = Dictionary_Get_Frame_Size();
          tag = Grammar_Tag(residue);
          switch (tag)
            { case NUL_tag:
                body = Stack_Pop();
                compiled_body = compile_sequence_C(body,
                                                   Pool_Lambda_Text);
                raw_frame_size = Dictionary_Exit_Nested_Scope();
                SURVIVE_DEFAULT_C(compiled_body);
                param_count = make_NBR(raw_param_count);
                frame_size = make_NBR(raw_frame_size);
                compiled_lambda = make_LMB_M(param_count,
                                             frame_size,
                                             compiled_body);
                return compiled_lambda;
              case SYM_tag:
                SURVIVE_DEFAULT_C(residue);
                Dictionary_Define_Symbol_M(residue,
                                           FALSE);
                body = Stack_Pop();
                compiled_body = compile_sequence_C(body,
                                                   Pool_Lambda_Text);
                raw_frame_size  = Dictionary_Exit_Nested_Scope();
                SURVIVE_DEFAULT_C(compiled_body);
                param_count = make_NBR(raw_param_count);
                frame_size = make_NBR(raw_frame_size);
                compiled_lambda_vararg = make_LMZ_M(param_count,
                                                    frame_size,
                                                    compiled_body);
                return compiled_lambda_vararg;
              default:
                message = IPA_message;
                goto error; }
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_Lambda_Text); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- literal ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_literal(EXP_type Literal)
  { return Literal; }

/*----------------------------------------------------------------------------*/
/*----------------------------------- load -----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_load_C(PAI_type Operands)
  { EXP_type compiled_expression,
             expression,
             residue;
    STR_type string;
    TAG_type tag;
    TXT_type message,
             input,
             path;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          message = E1X_message;
          goto error;
        case PAI_tag:
          string  = Operands->car;
          residue = Operands->cdr;
          if (!is_NUL(residue))
            { message = TMX_message;
              goto error; }
          if (!is_STR(string))
            { message = PNR_message;
              goto error; }
          path = string->txt;
          input = Main_Load(path);
          expression = Read_C(input);
          compiled_expression = compile_expression_C(expression);
          return compiled_expression;
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_Load_Text); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- parameters --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_parameters_C(PAI_type Parameters,
                                     TXT_type Error_text)
  { EXP_type parameter;
    PAI_type parameters;
    TXT_type message;
    for (parameters = Parameters;
         is_PAI(parameters);
         parameters = parameters->cdr)
      { SURVIVE_DEFAULT_C(parameters);
        parameter = parameters->car;
        if (!is_SYM(parameter))
          { message = IPA_message;
            goto error; }
        Dictionary_Define_Symbol_M(parameter,
                                   FALSE); }
    return parameters;
  error:
    return error_text(message,
                      Error_text); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- quote ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_quote_C(PAI_type Operands)
  { EXP_type expression,
             residue;
    QUO_type compiled_quote;
    TAG_type tag;
    TXT_type message;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          message = E1X_message;
          goto error;
        case PAI_tag:
          expression = Operands->car;
          residue    = Operands->cdr;
          if (!is_NUL(residue))
            { message = TMX_message;
              goto error; }
          SURVIVE_DEFAULT_C(expression);
          compiled_quote = make_QUO_M(expression);
          return compiled_quote;
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_Quote_Text); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- sequence ---------------------------------*/
/*----------------------------------------------------------------------------*/

static VEC_type compile_sequence_C(PAI_type Sequence,
                                   TXT_type Error_text)
  { EXP_type compiled_expression,
             expression;
    PAI_type sequence;
    UNS_type count;
    VEC_type compiled_sequence;
    sequence = Sequence;
    count = 0;
    do
      { sequence = Stack_Push_C(sequence);
        expression = sequence->car;
        compiled_expression = compile_expression_C(expression);
        sequence = Stack_Peek();
        sequence = sequence->cdr;
        Stack_Poke(compiled_expression);
        ++count; }
    while (is_PAI(sequence));
    if (!is_NUL(sequence))
      return error_text(ITF_message,
                        Error_text);
    Grammar_Dynamic_Claim_C(count);
    compiled_sequence = make_VEC_M(count);
    for (;
         count > 0;
         count--)
      { compiled_expression = Stack_Pop();
        compiled_sequence[count] = compiled_expression; }
    return compiled_sequence; }

/*----------------------------------------------------------------------------*/
/*------------------------------------ set -----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_set_C(PAI_type Operands)
  { EXP_type compiled_expression,
             expression;
    NBR_type offset,
             scope;
    PAI_type expressions,
             operands,
             residue;
    STL_type compiled_set_local;
    STG_type compiled_set_global;
    SYM_type variable;
    TAG_type tag;
    TXT_type message;
    UNS_type raw_offset,
             raw_scope;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          message = MSV_message;
          goto error;
        case PAI_tag:
          variable = Operands->car;
          if (is_SYM(variable))
            { expressions = Operands->cdr;
              tag = Grammar_Tag(expressions);
              switch (tag)
                { case NUL_tag:
                    message = E1X_message;
                    goto error;
                  case PAI_tag:
                    residue = expressions->cdr;
                    if (is_NUL(residue))
                      { operands = Stack_Push_C(Operands);
                        expressions = operands->cdr;
                        expression = expressions->car;
                        compiled_expression = compile_expression_C(expression);
                        operands = Stack_Pop();
                        variable = operands->car;
                        Dictionary_Lookup_Symbol(variable,
                                                 &raw_offset,
                                                 &raw_scope);
                        SURVIVE_DEFAULT_C(compiled_expression);
                        offset = make_NBR(raw_offset);
                        if (raw_scope == 0)
                          { compiled_set_local = make_STL_M(offset,
                                                            compiled_expression);
                            return compiled_set_local; }
                        else
                          { scope = make_NBR(raw_scope);
                            compiled_set_global = make_STG_M(scope,
                                                             offset,
                                                             compiled_expression);
                            return compiled_set_global; }}
                      else
                        { message = TMX_message;
                          goto error; }
                  default:
                    message = ITF_message;
                    goto error; }}
            else
              { message = IVV_message;
                goto error; }
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_Set_Text); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- symbol ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_symbol_C(SYM_type Variable_symbol)
  { GLB_type compiled_global_variable;
    LCL_type compiled_local_variable;
    NBR_type offset,
             scope;
    UNS_type raw_offset,
             raw_scope;
    Dictionary_Lookup_Symbol(Variable_symbol,
                             &raw_offset,
                             &raw_scope);
    Grammar_Default_Claim_C();
    if (raw_scope == 0)
      { offset = make_NBR(raw_offset);
        compiled_local_variable = make_LCL_M(offset);
        return compiled_local_variable; }
    scope  = make_NBR(raw_scope);
    offset = make_NBR(raw_offset);
    compiled_global_variable = make_GLB_M(scope,
                                          offset);
    return compiled_global_variable; }

/*----------------------------------------------------------------------------*/
/*----------------------------------- while ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_while_C(PAI_type Operands)
  { EXP_type body,
             compiled_body,
             compiled_predicate,
             predicate;
    NBR_type body_size,
             predicate_size;
    PAI_type operands;
    TAG_type tag;
    TXT_type message;
    UNS_type raw_body_size,
             raw_predicate_size;
    WHI_type compiled_while;
    tag = Grammar_Tag(Operands);
    switch (tag)
      { case NUL_tag:
          message = TFX_message;
          goto error;
        case PAI_tag:
          body = Operands->cdr;
          tag = Grammar_Tag(body);
          switch (tag)
            { case NUL_tag:
                message = TFX_message;
                goto error;
              case PAI_tag:
                SURVIVE_DEFAULT_C(Operands);
                Dictionary_Enter_Nested_Scope_M();
                operands = Stack_Push_C(Operands);
                predicate = operands->car;
                compiled_predicate = compile_expression_C(predicate);
                raw_predicate_size = Dictionary_Exit_Nested_Scope();
                SURVIVE_DEFAULT_C(compiled_predicate);
                Dictionary_Enter_Nested_Scope_M();
                operands = Stack_Peek();
                Stack_Poke(compiled_predicate);
                body = operands->cdr;
                compiled_body = compile_sequence_C(body,
                                                   Pool_While_Text);
                SURVIVE_DEFAULT_C(compiled_body);
                compiled_predicate = Stack_Pop();
                raw_body_size = Dictionary_Exit_Nested_Scope();
                predicate_size = make_NBR(raw_predicate_size);
                body_size = make_NBR(raw_body_size);
                compiled_while = make_WHI_M(compiled_predicate,
                                            predicate_size,
                                            compiled_body,
                                            body_size);
                return compiled_while;
              default:
                message = ITF_message;
                goto error; }
        default:
          message = ITF_message; }
  error:
    return error_text(message,
                      Pool_While_Text); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- form -----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_form_C(PAI_type Form)
  { EXP_type operands,
             operator;
    operator = Form->car;
    operands = Form->cdr;
    if (operator == Pool_Begin_P)
      return compile_begin_C(operands);
    if (operator == Pool_Define_P)
      return compile_define_C(operands);
    if (operator == Pool_If_P)
      return compile_if_C(operands);
    if (operator == Pool_Lambda_P)
      return compile_lambda_C(operands);
    if (operator == Pool_Load_P)
      return compile_load_C(operands);
    if (operator == Pool_Quote_P)
      return compile_quote_C(operands);
    if (operator == Pool_Set_P)
      return compile_set_C(operands);
    if (operator == Pool_While_P)
      return compile_while_C(operands);
    return compile_application_C(Form); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- expression --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type compile_expression_C(EXP_type Expression)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case PAI_tag:
          return compile_form_C(Expression);
        case SYM_tag:
          return compile_symbol_C(Expression);
        case CHA_tag:
        case FLS_tag:
        case NBR_tag:
        case NUL_tag:
        case REA_tag:
        case STR_tag:
        case TRU_tag:
        case VEC_tag:
          return compile_literal(Expression);
        default:
          return error_tag(IXT_message,
                           tag); }}

/*----------------------------- public functions -----------------------------*/

EXP_type Compile_C(EXP_type Expression)
  { EXP_type compiled_expression;
    compiled_expression = compile_expression_C(Expression);
    return compiled_expression; }

/*------------------------------ initialization ------------------------------*/

VOI_type Compile_Initialize(VOI_type)
  { return; }