                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 9: garbage collection   */
                     /*-----------------------------------*/
                     /*            dictionary             */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipGrammar.h"

#include "SlipDictionary.h"

/*----------------------------- private constants ----------------------------*/

static const TXT_type DPV_message = "duplicate variable name";
static const TXT_type TMG_message = "too many global variables";
static const TXT_type VNF_message = "variable not found";

/*----------------------------------- VAR ------------------------------------*/

#define NULL_VARIABLE (VAR_type)Grammar_Null

enum { VAR_size = 2 };

typedef struct VAR * VAR_type;

typedef struct VAR { CEL_type hdr;
                     SYM_type sym;
                     VAR_type var; } VAR;

static VAR_type make_VAR_M(SYM_type Symbol,
                           VAR_type Variable)
  { VAR_type variable = (VAR_type)make_VEC_M(VAR_size);
    variable->sym = Symbol;
    variable->var = Variable;
    return variable; }

/*----------------------------------- DCT ------------------------------------*/

#define NULL_DICTIONARY (DCT_type)Grammar_Null

enum { DCT_size = 3 };

typedef struct DCT * DCT_type;

typedef struct DCT { CEL_type hdr;
                     VAR_type var;
                     NBR_type cnt;
                     DCT_type dct; } DCT;

static DCT_type make_DCT_M(VAR_type Variable,
                           NBR_type Count,
                           DCT_type Dictionary)
  { DCT_type dictionary = (DCT_type)make_VEC_M(DCT_size);
    dictionary->var = Variable;
    dictionary->cnt = Count;
    dictionary->dct = Dictionary;
    return dictionary; }

/*----------------------------- private variables ----------------------------*/

static DCT_type Current_dictionary_P;
static VAR_type Current_variable_P;
static VAR_type Previous_variable_P;

/*----------------------------------------------------------------------------*/

static UNS_type Current_nesting_level;
static UNS_type Current_variable_count;
static UNS_type Previous_variable_count;

/*----------------------------- private functions ----------------------------*/

static VOI_type error(TXT_type Message)
  { Main_Error(Message); }

static VOI_type error_symbol(TXT_type Message,
                             SYM_type Symbol)
  { TXT_type text;
    text = Symbol->txt;
    sprintf(Main_Buffer, "%s: %s", Message, text);
    error(Main_Buffer); }

static UNS_type get_raw_offset(SYM_type Symbol,
                               UNS_type Count,
                               VAR_type Variable)
  { VAR_type variable;
    UNS_type count = Count;
    for (variable = Variable;
         !is_NUL(variable);
         variable = variable->var)
      { if (variable->sym == Symbol)
          return count;
        count--; }
    return 0; }

/*----------------------------- public functions -----------------------------*/

VOI_type Dictionary_Checkpoint(VOI_type)
  { Previous_variable_P = Current_variable_P;
    Previous_variable_count = Current_variable_count; }

UNS_type Dictionary_Define_Symbol_M(SYM_type Symbol,
                                    BLN_type Duplicates)
  { UNS_type raw_offset;
    if ((Current_nesting_level == 0) &&
        (Current_variable_count >= Grammar_Global_Frame_Size))
      error_symbol(TMG_message,
                   Symbol);
    raw_offset = get_raw_offset(Symbol,
                                Current_variable_count,
                                Current_variable_P);
    if (raw_offset != 0)
      { if (Duplicates)
          return raw_offset;
        error_symbol(DPV_message,
                     Symbol); }
    Current_variable_P = make_VAR_M(Symbol,
                                    Current_variable_P);
    return ++Current_variable_count; }

VOI_type Dictionary_Enter_Nested_Scope_M(VOI_type)
  { NBR_type variable_count;
    variable_count = make_NBR(Current_variable_count);
    Current_dictionary_P = make_DCT_M(Current_variable_P,
                                      variable_count,
                                      Current_dictionary_P);
    Current_variable_P = NULL_VARIABLE;
    Current_variable_count = 0;
    Current_nesting_level++; }

UNS_type Dictionary_Exit_Nested_Scope(VOI_type)
  { NBR_type variable_count;
    UNS_type previous_raw_variable_count;
    previous_raw_variable_count = Current_variable_count;
    variable_count       = Current_dictionary_P->cnt;
    Current_variable_P   = Current_dictionary_P->var;
    Current_dictionary_P = Current_dictionary_P->dct;
    Current_nesting_level--;
    Current_variable_count = get_NBR(variable_count);
    return previous_raw_variable_count; }

UNS_type Dictionary_Get_Frame_Size(VOI_type)
  { return Current_variable_count; }

VOI_type Dictionary_Lookup_Symbol(SYM_type Symbol,
                                  URF_type Offset_reference,
                                  URF_type Scope_reference)
  { DCT_type dictionary;
    VAR_type variable;
    NBR_type variable_count;
    UNS_type raw_variable_count,
             raw_offset,
             raw_scope;
    raw_offset = get_raw_offset(Symbol,
                                Current_variable_count,
                                Current_variable_P);
    if (raw_offset > 0)
      { *Scope_reference = 0;
        *Offset_reference = raw_offset;
        return; }
    raw_scope = Current_nesting_level;
    for (dictionary = Current_dictionary_P;
         !is_NUL(dictionary);
         dictionary = dictionary->dct)
      { variable_count = dictionary->cnt;
        variable       = dictionary->var;
        raw_variable_count = get_NBR(variable_count);
        raw_offset = get_raw_offset(Symbol,
                                    raw_variable_count,
                                    variable);
        if (raw_offset > 0)
          { *Scope_reference = raw_scope;
            *Offset_reference = raw_offset;
            return; }
        raw_scope--; }
    error_symbol(VNF_message,
                 Symbol); }

VOI_type Dictionary_Rollback(VOI_type)
  { Current_variable_P = Previous_variable_P;
    Current_variable_count = Previous_variable_count;
    Current_dictionary_P = NULL_DICTIONARY;
    Current_nesting_level = 0; }

/*------------------------------ initialization ------------------------------*/

VOI_type Dictionary_Initialize(VOI_type)
  { Current_dictionary_P = NULL_DICTIONARY;
    Current_nesting_level = 0;
    Previous_variable_P = Current_variable_P = NULL_VARIABLE;
    Previous_variable_count = Current_variable_count = 0;
    REGISTER(Current_dictionary_P);
    REGISTER(Current_variable_P);
    REGISTER(Previous_variable_P); }