                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 9: garbage collection   */
                     /*-----------------------------------*/
                     /*              evaluate             */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipGrammar.h"

#include "SlipEnvironment.h"
#include "SlipPool.h"
#include "SlipStack.h"
#include "SlipThread.h"

#include "SlipEvaluate.h"

/*------------------------------- private macros -----------------------------*/

#define NULL_PAIR (PAI_type)Grammar_Null

/*----------------------------- private constants ----------------------------*/

static const TXT_type BRQ_message = "boolean value required in if";
static const TXT_type IFS_message = "illegal frame size";
static const TXT_type IXT_message = "invalid expression type";
static const TXT_type PNR_message = "procedure or native required";
static const TXT_type TFO_message = "too few operands in application";
static const TXT_type TMO_message = "too many operands in application";

/*---------------------------- private prototypes ----------------------------*/

static EXP_type         evaluate_bindings_C(PRC_type);
static EXP_type        evaluate_bindings_DC(VEC_type);
static EXP_type        evaluate_body_poke_M(VEC_type,
                                            VEC_type,
                                            VEC_type);
static EXP_type        evaluate_body_push_M(VEC_type,
                                            VEC_type,
                                            VEC_type);
static EXP_type       evaluate_expression_C(EXP_type);
static EXP_type      evaluate_native_call_C(NAT_type);
static EXP_type     evaluate_native_call_DC(VEC_type);
static EXP_type         evaluate_sequence_C(VEC_type);
static EXP_type  evaluate_vararg_bindings_C(PRZ_type);
static EXP_type evaluate_vararg_bindings_DC(VEC_type);

/*----------------------------- private functions ----------------------------*/

static EXP_type error(TXT_type Message)
  { Main_Error(Message);
    return Grammar_Unspecified; }

static EXP_type error_procedure(TXT_type Message,
                                PRC_type Procedure)
  { SYM_type name;
    TXT_type text;
    name = Procedure->nam;
    text = name->txt;
    sprintf(Main_Buffer, "%s: %s", Message, text);
    return error(Main_Buffer); }

static EXP_type error_procedure_vararg(TXT_type Message,
                                       PRZ_type Procedure)
  { SYM_type name;
    TXT_type text;
    name = Procedure->nam;
    text = name->txt;
    sprintf(Main_Buffer, "%s: %s", Message, text);
    return error(Main_Buffer); }

static EXP_type error_tag(TXT_type Message,
                          TAG_type Tag)
  { sprintf(Main_Buffer, "%s: %d", Message, Tag);
    return error(Main_Buffer); }

static EXP_type error_text(TXT_type Message,
                           TXT_type Text)
  { sprintf(Main_Buffer, "%s: %s", Message, Text);
    return error(Main_Buffer); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- application -------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_application;

BEGIN_THREAD_FRAME(aPL);
 THREAD_FRAME_SLOT(VEC, opd);
END_THREAD_FRAME(aPL);

static EXP_type continue_application_C(EXP_type Procedure)
  { TAG_type tag;
    tag = Grammar_Tag(Procedure);
    switch (tag)
      { case NAT_tag:
          return evaluate_native_call_C(Procedure);
        case PRC_tag:
          return evaluate_bindings_C(Procedure);
        case PRZ_tag:
          return evaluate_vararg_bindings_C(Procedure);
        default:
          return error_text(PNR_message,
                            Pool_Application_Text); }}

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_application_C(APL_type Application)
  { aPL_type application_thread;
    EXP_type expression;
    VEC_type operands;
    SURVIVE_DEFAULT_C(Application);
    expression = Application->exp;
    operands   = Application->opr;
    application_thread = (aPL_type)Thread_Push_M(Continue_application,
                                                 aPL_size);
    application_thread->opd = operands;
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_application_DC(VEC_type Frame)
  { EXP_type value;
    TAG_type tag;
    value = Frame[1];
    tag = Grammar_Tag(value);
    switch (tag)
      { case NAT_tag:
          return evaluate_native_call_DC(Frame);
        case PRC_tag:
          return evaluate_bindings_DC(Frame);
        case PRZ_tag:
          return evaluate_vararg_bindings_DC(Frame);
        default:
          return error_text(PNR_message,
                            Pool_Application_Text); }}

/*----------------------------------------------------------------------------*/

static VOI_type initialize_application(VOI_type)
  { Continue_application = Thread_Register(continue_application_C); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- begin ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_begin_C(BEG_type Begin)
  { VEC_type sequence;
    sequence = Begin->seq;
    return evaluate_sequence_C(sequence); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- bindings --------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_bindings;

BEGIN_THREAD_FRAME(bND);
 THREAD_FRAME_SLOT(PRC, prc);
 THREAD_FRAME_SLOT(VEC, opd);
 THREAD_FRAME_SLOT(VEC, frm);
 THREAD_FRAME_SLOT(NBR, pos);
END_THREAD_FRAME(bND);

static EXP_type continue_bindings_C(EXP_type Value)
  { bND_type binding_thread;
    EXP_type expression;
    NBR_type param_count,
             position;
    PRC_type procedure;
    UNS_type index,
             raw_param_count;
    VEC_type body,
             environment,
             frame,
             operands;
    SURVIVE_DEFAULT_C(Value);
    binding_thread = (bND_type)Thread_Peek();
    procedure = binding_thread->prc;
    frame     = binding_thread->frm;
    position  = binding_thread->pos;
    index = get_NBR(position);
    frame[index] = Value;
    param_count = procedure->par;
    raw_param_count = get_NBR(param_count);
    if (index < raw_param_count)
      { binding_thread = (bND_type)Thread_Keep();
        position = make_NBR(++index);
        operands = binding_thread->opd;
        binding_thread->pos = position;
        expression = operands[index];
        return evaluate_expression_C(expression); }
    body        = procedure->bod;
    environment = procedure->env;
    return evaluate_body_poke_M(body,
                                environment,
                                frame); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_bindings_C(PRC_type Procedure)
  { aPL_type application_thread;
    bND_type binding_thread;
    EXP_type expression;
    NBR_type frame_size,
             param_count;
    UNS_type raw_frame_size,
             raw_operand_count,
             raw_param_count;
    VEC_type body,
             environment,
             frame,
             operands;
    application_thread = (aPL_type)Thread_Peek();
    operands = application_thread->opd;
    param_count = Procedure->par;
    raw_param_count = get_NBR(param_count);
    raw_operand_count = size_VEC(operands);
    if (raw_param_count > raw_operand_count)
      return error_procedure(TFO_message,
                             Procedure);
    if (raw_param_count < raw_operand_count)
      return error_procedure(TMO_message,
                             Procedure);
    frame_size = Procedure->siz;
    raw_frame_size = get_NBR(frame_size);
    if (raw_frame_size == 0)
      { SURVIVE_DEFAULT_C(Procedure);
        body        = Procedure->bod;
        environment = Procedure->env;
        frame = Grammar_Empty_Vector_P;
        return evaluate_body_poke_M(body,
                                    environment,
                                    frame); }
    SURVIVE_DYNAMIC_C(Procedure,
                      raw_frame_size);
    frame = Environment_Create_Frame_M(raw_frame_size);
    if (raw_param_count == 0)
      { body        = Procedure->bod;
        environment = Procedure->env;
        return evaluate_body_poke_M(body,
                                    environment,
                                    frame); }
    application_thread = (aPL_type)Thread_Peek();
    operands = application_thread->opd;
    binding_thread = (bND_type)Thread_Poke_M(Continue_bindings,
                                             bND_size);
    binding_thread->prc = Procedure;
    binding_thread->opd = operands;
    binding_thread->frm = frame;
    binding_thread->pos = Grammar_One;
    expression = operands[1];
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_bindings_DC(VEC_type Frame)
  { EXP_type value;
    NBR_type frame_size,
             param_count;
    PAI_type arguments;
    PRC_type procedure;
    UNS_type raw_frame_size,
             raw_param_count,
             position;
    VEC_type body,
             environment,
             frame;
    procedure = Frame[1];
    frame_size = procedure->siz;
    raw_frame_size = get_NBR(frame_size);
    SURVIVE_DYNAMIC_C(Frame,
                      raw_frame_size);
    procedure = Frame[1];
    arguments = Frame[2];
    param_count = procedure->par;
    body        = procedure->bod;
    environment = procedure->env;
    raw_param_count = get_NBR(param_count);
    if (raw_frame_size == 0)
      frame = Grammar_Empty_Vector_P;
    else
      { frame = Environment_Create_Frame_M(raw_frame_size);
        for (position = 1;
             position <= raw_param_count;
             position++)
          { if (!is_PAI(arguments))
              return error_procedure(TFO_message,
                                     procedure);
            value     = arguments->car;
            arguments = arguments->cdr;
            frame[position] = value; }}
    if (!is_NUL(arguments))
      return error_procedure(TMO_message,
                             procedure);
    return evaluate_body_push_M(body,
                                environment,
                                frame); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_bindings(VOI_type)
  { Continue_bindings = Thread_Register(continue_bindings_C); }

/*----------------------------------------------------------------------------*/
/*----------------------------------- body -----------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_body;

BEGIN_THREAD_FRAME(bOD);
 THREAD_FRAME_SLOT(VEC, env);
 THREAD_FRAME_SLOT(VEC, frm);
END_THREAD_FRAME(bOD);

static EXP_type continue_body(EXP_type Value)
  { bOD_type body_thread;
    VEC_type environment,
             frame;
    body_thread = (bOD_type)Thread_Peek();
    environment = body_thread->env;
    frame       = body_thread->frm;
    Thread_Zap();
    Environment_Set_Environment_And_Frame(environment,
                                          frame);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_body_poke_M(VEC_type Body,
                                     VEC_type Environment,
                                     VEC_type Frame)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Poke_M(Continue_body,
                                          bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(Environment,
                                          Frame);
    return evaluate_sequence_C(Body); }

static EXP_type evaluate_body_push_M(VEC_type Body,
                                     VEC_type Environment,
                                     VEC_type Frame)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Push_M(Continue_body,
                                          bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(Environment,
                                          Frame);
    return evaluate_sequence_C(Body); }

static EXP_type evaluate_inline_poke_C(EXP_type Expression,
                                       UNS_type Frame_size)
  { bOD_type body_thread;
    UNS_type environment_size;
    VEC_type environment,
             frame;
    environment_size = Environment_Get_Environment_size();
    SURVIVE_DYNAMIC_C(Expression,
                      environment_size + Frame_size);
    environment = Environment_Grow_Environment_M();
    frame = Environment_Create_Frame_M(Frame_size);
    body_thread = (bOD_type)Thread_Poke_M(Continue_body,
                                          bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(environment,
                                          frame);
    return evaluate_expression_C(Expression); }

static EXP_type evaluate_inline_push_C(EXP_type Expression,
                                       UNS_type Frame_size)
  { bOD_type body_thread;
    UNS_type environment_size;
    VEC_type environment,
             frame;
    environment_size = Environment_Get_Environment_size();
    SURVIVE_DYNAMIC_C(Expression,
                      environment_size + Frame_size);
    environment = Environment_Grow_Environment_M();
    frame = Environment_Create_Frame_M(Frame_size);
    body_thread = (bOD_type)Thread_Push_M(Continue_body,
                                          bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(environment,
                                          frame);
    return evaluate_expression_C(Expression); }

static EXP_type evaluate_inline_sequence_C(VEC_type Body,
                                           UNS_type Frame_size)
  { bOD_type body_thread;
    UNS_type environment_size;
    VEC_type environment,
             frame;
    environment_size = Environment_Get_Environment_size();
    SURVIVE_DYNAMIC_C(Body,
                      environment_size + Frame_size);
    environment = Environment_Grow_Environment_M();
    frame = Environment_Create_Frame_M(Frame_size);
    body_thread = (bOD_type)Thread_Push_M(Continue_body,
                                          bOD_size);
    body_thread->env = Environment_Get_Environment();
    body_thread->frm = Environment_Get_Frame();
    Environment_Set_Environment_And_Frame(environment,
                                          frame);
    return evaluate_sequence_C(Body); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_body(VOI_type)
  { Continue_body = Thread_Register(continue_body); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- define ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_define;

BEGIN_THREAD_FRAME(dEF);
 THREAD_FRAME_SLOT(NBR, ofs);
END_THREAD_FRAME(dEF);

static EXP_type continue_define(EXP_type Value)
  { dEF_type define_thread;
    NBR_type offset;
    UNS_type raw_offset;
    define_thread = (dEF_type)Thread_Peek();
    offset = define_thread->ofs;
    Thread_Zap();
    raw_offset = get_NBR(offset);
    Environment_Local_Set(raw_offset,
                          Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_define_function_C(DFF_type Define)
  { NBR_type frame_size,
             offset,
             param_count;
    PRC_type procedure;
    SYM_type name;
    UNS_type environment_size,
             raw_offset;
    VEC_type body,
             environment;
    environment_size = Environment_Get_Environment_size();
    SURVIVE_DYNAMIC_C(Define,
                      environment_size);
    name        = Define->nam;
    offset      = Define->ofs;
    param_count = Define->par;
    frame_size  = Define->siz;
    body        = Define->bod;
    environment = Environment_Grow_Environment_M();
    procedure = make_PRC_M(name,
                           param_count,
                           frame_size,
                           body,
                           environment);
    raw_offset = get_NBR(offset);
    Environment_Local_Set(raw_offset,
                          procedure);
    return procedure; }

static EXP_type evaluate_define_function_vararg_C(DFZ_type Define)
  { NBR_type frame_size,
             offset,
             param_count;
    PRZ_type procedure;
    SYM_type name;
    UNS_type environment_size,
             raw_offset;
    VEC_type body,
             environment;
    environment_size = Environment_Get_Environment_size();
    SURVIVE_DYNAMIC_C(Define,
                      environment_size);
    name        = Define->nam;
    offset      = Define->ofs;
    param_count = Define->par;
    frame_size  = Define->siz;
    body        = Define->bod;
    environment = Environment_Grow_Environment_M();
    procedure = make_PRZ_M(name,
                           param_count,
                           frame_size,
                           body,
                           environment);
    raw_offset = get_NBR(offset);
    Environment_Local_Set(raw_offset,
                          procedure);
    return procedure; }

static EXP_type evaluate_define_variable_C(DFV_type Define)
  { dEF_type define_thread;
    EXP_type expression;
    NBR_type offset;
    SURVIVE_DEFAULT_C(Define);
    offset     = Define->ofs;
    expression = Define->exp;
    define_thread = (dEF_type)Thread_Push_M(Continue_define,
                                            dEF_size);
    define_thread->ofs = offset;
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_define(VOI_type)
  { Continue_define = Thread_Register(continue_define); }

/*----------------------------------------------------------------------------*/
/*------------------------------ global variable -----------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_global_variable(GLB_type Global)
  { EXP_type value;
    NBR_type offset,
             scope;
    UNS_type raw_scope,
             raw_offset;
    scope  = Global->scp;
    offset = Global->ofs;
    raw_scope = get_NBR(scope);
    raw_offset = get_NBR(offset);
    value = Environment_Global_Get(raw_scope,
                                   raw_offset);
    return value; }

/*----------------------------------------------------------------------------*/
/*------------------------------------ if ------------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_if_double;

BEGIN_THREAD_FRAME(iFD);
 THREAD_FRAME_SLOT(EXP, cns);
 THREAD_FRAME_SLOT(NBR, csz);
 THREAD_FRAME_SLOT(EXP, alt);
 THREAD_FRAME_SLOT(NBR, asz);
END_THREAD_FRAME(iFD);

static EXP_type continue_if_double_C(EXP_type Boolean)
  { iFD_type if_thread;
    EXP_type alternative,
             consequent;
    NBR_type alternative_size,
             consequent_size;
    TAG_type tag;
    UNS_type raw_alternative_size,
             raw_consequent_size;
    tag = Grammar_Tag(Boolean);
    if_thread = (iFD_type)Thread_Peek();
    switch (tag)
      { case FLS_tag:
          alternative      = if_thread->alt;
          alternative_size = if_thread->asz;
          raw_alternative_size = get_NBR(alternative_size);
          return evaluate_inline_poke_C(alternative,
                                        raw_alternative_size);
        case TRU_tag:
          consequent      = if_thread->cns;
          consequent_size = if_thread->csz;
          raw_consequent_size = get_NBR(consequent_size);
          return evaluate_inline_poke_C(consequent,
                                        raw_consequent_size);
        default:
          return error(BRQ_message); }}

/*----------------------------------------------------------------------------*/

static NBR_type Continue_if_single;

BEGIN_THREAD_FRAME(iFS);
 THREAD_FRAME_SLOT(EXP, cns);
 THREAD_FRAME_SLOT(NBR, csz);
END_THREAD_FRAME(iFS);

static EXP_type continue_if_single_C(EXP_type Boolean)
  { iFS_type if_thread;
    EXP_type consequent;
    NBR_type consequent_size;
    TAG_type tag;
    UNS_type raw_consequent_size;
    tag = Grammar_Tag(Boolean);
    if_thread = (iFS_type)Thread_Peek();
    switch (tag)
      { case FLS_tag:
          Thread_Zap();
          return Grammar_Unspecified;
        case TRU_tag:
          consequent      = if_thread->cns;
          consequent_size = if_thread->csz;
          raw_consequent_size = get_NBR(consequent_size);
          return evaluate_inline_poke_C(consequent,
                                        raw_consequent_size);
        default:
          return error(BRQ_message); }}

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_if_double_C(IFF_type If)
  { iFD_type if_thread;
    EXP_type alternative,
             consequent,
             predicate;
    NBR_type alternative_size,
             consequent_size;
    SURVIVE_DEFAULT_C(If);
    predicate        = If->prd;
    consequent       = If->cns;
    consequent_size  = If->csz;
    alternative      = If->alt;
    alternative_size = If->asz;
    if_thread = (iFD_type)Thread_Push_M(Continue_if_double,
                                        iFD_size);
    if_thread->cns = consequent;
    if_thread->csz = consequent_size;
    if_thread->alt = alternative;
    if_thread->asz = alternative_size;
    return evaluate_expression_C(predicate); }

static EXP_type evaluate_if_single_C(IFZ_type If)
  { iFS_type if_thread;
    EXP_type consequent,
             predicate;
    NBR_type consequent_size;
    SURVIVE_DEFAULT_C(If);
    predicate       = If->prd;
    consequent      = If->cns;
    consequent_size = If->csz;
    if_thread = (iFS_type)Thread_Push_M(Continue_if_single,
                                        iFS_size);
    if_thread->cns = consequent;
    if_thread->csz = consequent_size;
    return evaluate_expression_C(predicate); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_if(VOI_type)
  { Continue_if_double = Thread_Register(continue_if_double_C);
    Continue_if_single = Thread_Register(continue_if_single_C); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- lambda ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_lambda_C(LMB_type Lambda)
  { NBR_type frame_size,
             param_count;
    PRC_type procedure;
    UNS_type environment_size;
    VEC_type body,
             environment;
    environment_size = Environment_Get_Environment_size();
    SURVIVE_DYNAMIC_C(Lambda,
                      environment_size);
    param_count = Lambda->par;
    frame_size  = Lambda->siz;
    body        = Lambda->bod;
    environment = Environment_Grow_Environment_M();
    procedure = make_PRC_M(Pool_Anonymous_P,
                           param_count,
                           frame_size,
                           body,
                           environment);
    return procedure; }

static EXP_type evaluate_lambda_vararg_C(LMZ_type Lambda)

  { NBR_type frame_size,
             param_count;
    PRZ_type procedure;
    UNS_type environment_size;
    VEC_type body,
             environment;
    environment_size = Environment_Get_Environment_size();
    SURVIVE_DYNAMIC_C(Lambda,
                      environment_size);
    param_count = Lambda->par;
    frame_size  = Lambda->siz;
    body        = Lambda->bod;
    environment = Environment_Grow_Environment_M();
    procedure = make_PRZ_M(Pool_Anonymous_P,
                           param_count,
                           frame_size,
                           body,
                           environment);
    return procedure; }

/*----------------------------------------------------------------------------*/
/*--------------------------------- literal ----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type clone_C(EXP_type);

static PAI_type clone_pair_C(PAI_type Pair)
  { EXP_type car_value,
             cdr_value;
    PAI_type pair;
    SURVIVE_DEFAULT_C(Pair);
    car_value = Pair->car;
    cdr_value = Pair->cdr;
    pair = make_PAI_M(car_value,
                      cdr_value);
    pair = Stack_Push_C(pair);
    car_value = pair->car;
    car_value = clone_C(car_value);
    pair = Stack_Peek();
    pair->car = car_value;
    cdr_value = pair->cdr;
    cdr_value = clone_C(cdr_value);
    pair = Stack_Pop();
    pair->cdr = cdr_value;
    return pair; }

static STR_type clone_string_C(STR_type String)
  { STR_type cloned_string;
    TXT_type text;
    UNS_type size;
    text = String->txt;
    size = size_STR(text);
    SURVIVE_DYNAMIC_C(String,
                      size);
    text = String->txt;
    cloned_string = make_STR_M(text);
    return cloned_string; }

static VEC_type clone_vector_C(VEC_type Vector)
  { EXP_type value;
    UNS_type index,
             size;
    VEC_type vector;
    size = size_VEC(Vector);
    if (size == 0)
      return Grammar_Empty_Vector_P;
    SURVIVE_DYNAMIC_C(Vector,
                      size);
    vector = make_VEC_M(size);
    for (index = 1;
         index <= size;
         index += 1)
      vector[index] = Vector[index];
    vector = Stack_Push_C(vector);
    for (index = 1;
         index <= size;
         index += 1)
      { value = vector[index];
        value = clone_C(value);
        vector = Stack_Peek();
        vector[index] = value; }
    return vector; }

static EXP_type clone_C(EXP_type Literal)
  { TAG_type tag;
    tag = Grammar_Tag(Literal);
    switch (tag)
      { case PAI_tag:
          return clone_pair_C(Literal);
        case STR_tag:
          return clone_string_C(Literal);
        case VEC_tag:
          return clone_vector_C(Literal); 
        default:
          return Literal; }}

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_literal(EXP_type Literal)
  { return clone_C(Literal); }

/*----------------------------------------------------------------------------*/
/*------------------------------ local variable ------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_local_variable(LCL_type Local)
  { EXP_type value;
    NBR_type offset;
    UNS_type raw_offset;
    offset = Local->ofs;
    raw_offset = get_NBR(offset);
    value = Environment_Local_Get(raw_offset);
    return value; }

/*----------------------------------------------------------------------------*/
/*---------------------------------- native ----------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_native_call;

BEGIN_THREAD_FRAME(nCL);
 THREAD_FRAME_SLOT(VEC, opd);
 THREAD_FRAME_SLOT(VEC, frm);
 THREAD_FRAME_SLOT(NAT, nat);
 THREAD_FRAME_SLOT(NBR, pos);
END_THREAD_FRAME(nCL);

static EXP_type continue_native_call_C(EXP_type Value)
  { nCL_type native_call_thread;
    EXP_type expression;
    FUN_type function;
    NAT_type native;
    NBR_type position;
    UNS_type index,
             size_v;
    VEC_type frame,
             operands;
    native_call_thread = (nCL_type)Thread_Peek();
    frame    = native_call_thread->frm;
    position = native_call_thread->pos;
    index = get_NBR(position);
    size_v = size_VEC(frame);
    frame[index] = Value;
    if (index++ < size_v)
      { native_call_thread = (nCL_type)Thread_Keep();
        operands = native_call_thread->opd;
        expression = operands[index];
        position = make_NBR(index);
        native_call_thread->pos = position;
        return evaluate_expression_C(expression); }
    native = native_call_thread->nat;
    function = native->fun;
    Thread_Zap();
    return function(frame); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_native_call_C(NAT_type Native)
  { aPL_type application_thread;
    nCL_type native_call_thread;
    EXP_type expression;
    FUN_type function;
    UNS_type raw_operand_size;
    VEC_type frame,
             operands;
    application_thread = (aPL_type)Thread_Peek();
    operands = application_thread->opd;
    raw_operand_size = size_VEC(operands);
    if (raw_operand_size == 0)
      { function = Native->fun;
        Thread_Zap();
        return function(Grammar_Empty_Vector_P); }
    SURVIVE_DYNAMIC_C(Native,
                      raw_operand_size);
    application_thread = (aPL_type)Thread_Peek();
    operands = application_thread->opd;
    frame = Environment_Create_Frame_M(raw_operand_size);
    native_call_thread = (nCL_type)Thread_Poke_M(Continue_native_call,
                                                 nCL_size);
    native_call_thread->opd = operands;
    native_call_thread->frm = frame;
    native_call_thread->nat = Native;
    native_call_thread->pos = Grammar_One;
    expression = operands[1];
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_native_call_DC(VEC_type Frame)
  { EXP_type value;
    FUN_type function;
    NAT_type native;
    PAI_type arguments;
    TXT_type text;
    UNS_type length,
             position;
    VEC_type frame;
    native    = Frame[1];
    arguments = Frame[2];
    function = native->fun;
    text     = native->txt;
    length = Grammar_List_Length(arguments,
                                 text);
    if (length == 0)
      frame = Grammar_Empty_Vector_P;
    else
      { SURVIVE_DYNAMIC_C(arguments,
                          length);
        frame = Environment_Create_Frame_M(length);
        for (position = 1;
             position <= length;
             position++)
          { value     = arguments->car;
            arguments = arguments->cdr;
            frame[position] = value; }}
    return function(frame); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_native_call(VOI_type)
  { Continue_native_call = Thread_Register(continue_native_call_C); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- quote -----------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_quote_C(QUO_type Quote)
  { EXP_type expression;
    expression = Quote->exp;
    return clone_C(expression); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- sequence ---------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_sequence;

BEGIN_THREAD_FRAME(sEQ);
 THREAD_FRAME_SLOT(VEC, exs);
 THREAD_FRAME_SLOT(NBR, pos);
END_THREAD_FRAME(sEQ);

static EXP_type continue_sequence_C(EXP_type Ignore)
  { sEQ_type sequence_thread;
    EXP_type expression;
    VEC_type expressions;
    NBR_type position;
    UNS_type index,
             size_x;
    sequence_thread = (sEQ_type)Thread_Peek();
    expressions = sequence_thread->exs;
    position    = sequence_thread->pos;
    index = get_NBR(position);
    size_x = size_VEC(expressions);
    if (++index < size_x)
      { sequence_thread = (sEQ_type)Thread_Keep();
        position = make_NBR(index);
        sequence_thread->pos = position; }
    else
      Thread_Zap();
    expression = expressions[index];
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_sequence_C(VEC_type Expressions)
  { sEQ_type sequence_thread;
    EXP_type expression;
    UNS_type size_x;
    size_x = size_VEC(Expressions);
    if (size_x > 1)
      { SURVIVE_DEFAULT_C(Expressions);
        sequence_thread = (sEQ_type)Thread_Push_M(Continue_sequence,
                                                  sEQ_size);
        sequence_thread->exs = Expressions;
        sequence_thread->pos = Grammar_One; }
    expression = Expressions[1];
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_sequence(VOI_type)
  { Continue_sequence = Thread_Register(continue_sequence_C); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- set global --------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_set_global;

BEGIN_THREAD_FRAME(sTG);
 THREAD_FRAME_SLOT(NBR, scp);
 THREAD_FRAME_SLOT(NBR, ofs);
END_THREAD_FRAME(sTG);

static EXP_type continue_set_global(EXP_type Value)
  { sTG_type set_thread;
    NBR_type offset,
             scope;
    UNS_type raw_offset,
             raw_scope;
    set_thread = (sTG_type)Thread_Peek();
    scope  = set_thread->scp;
    offset = set_thread->ofs;
    Thread_Zap();
    raw_scope = get_NBR(scope);
    raw_offset = get_NBR(offset);
    Environment_Global_Set(raw_scope,
                           raw_offset,
                           Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_set_global_C(STG_type Set)
  { sTG_type set_thread;
    EXP_type expression;
    NBR_type offset,
             scope;
    SURVIVE_DEFAULT_C(Set);
    set_thread = (sTG_type)Thread_Push_M(Continue_set_global,
                                         sTG_size);
    scope  = Set->scp;
    offset = Set->ofs;
    expression = Set->exp;
    set_thread->scp = scope;
    set_thread->ofs = offset;
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_set_global(VOI_type)
  { Continue_set_global = Thread_Register(continue_set_global); }

/*----------------------------------------------------------------------------*/
/*--------------------------------- set local --------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_set_local;

BEGIN_THREAD_FRAME(sTL);
 THREAD_FRAME_SLOT(NBR, ofs);
END_THREAD_FRAME(sTL);

static EXP_type continue_set_local(EXP_type Value)
  { sTL_type set_thread;
    NBR_type offset;
    UNS_type raw_offset;
    set_thread = (sTL_type)Thread_Peek();
    offset = set_thread->ofs;
    Thread_Zap();
    raw_offset = get_NBR(offset);
    Environment_Local_Set(raw_offset,
                          Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_set_local_C(STL_type Set)
  { sTL_type set_thread;
    EXP_type expression;
    NBR_type offset;
    SURVIVE_DEFAULT_C(Set);
    set_thread = (sTL_type)Thread_Push_M(Continue_set_local,
                                         sTL_size);
    offset     = Set->ofs;
    expression = Set->exp;
    set_thread->ofs = offset;
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_set_local(VOI_type)
  { Continue_set_local = Thread_Register(continue_set_local); }

/*----------------------------------------------------------------------------*/
/*----------------------------- vararg bindings ------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_vararg,
                Continue_vararg_bindings;

BEGIN_THREAD_FRAME(vRG);
 THREAD_FRAME_SLOT(PRZ, prc);
 THREAD_FRAME_SLOT(VEC, frm);
 THREAD_FRAME_SLOT(VEC, opd);
 THREAD_FRAME_SLOT(PAI, lst);
 THREAD_FRAME_SLOT(NBR, pos);
END_THREAD_FRAME(vRG);

static EXP_type continue_vararg_C(EXP_type Value)
  { vRG_type binding_thread;
    EXP_type expression;
    NBR_type param_count,
             position;
    PAI_type list;
    PRZ_type procedure;
    SYM_type name;
    TXT_type text;
    UNS_type index,
             operand_size,
             raw_param_count;
    VEC_type body,
             environment,
             frame,
             operands;
    SURVIVE_DEFAULT_C(Value);
    binding_thread = (vRG_type)Thread_Peek();
    operands = binding_thread->opd;
    list     = binding_thread->lst;
    position = binding_thread->pos;
    index = get_NBR(position);
    operand_size = size_VEC(operands);
    list = make_PAI_M(Value,
                      list);
    if (index < operand_size)
      { binding_thread = (vRG_type)Thread_Keep();
        position = make_NBR(++index);
        binding_thread->lst = list;
        binding_thread->pos = position;
        expression = operands[index];
        return evaluate_expression_C(expression); }
    procedure = binding_thread->prc;
    frame     = binding_thread->frm;
    param_count = procedure->par;
    body        = procedure->bod;
    environment = procedure->env;
    raw_param_count = get_NBR(param_count);
    name = procedure->nam;
    text = name->txt;
    list = Grammar_Reverse_List(list,
                                text);
    frame[raw_param_count + 1] = list;
    return evaluate_body_poke_M(body,
                                environment,
                                frame); }

static EXP_type continue_vararg_bindings_C(EXP_type Value)
  { vRG_type binding_thread;
    EXP_type expression;
    NBR_type param_count,
             position;
    PRZ_type procedure;
    UNS_type index,
             raw_operand_count,
             raw_param_count;
    VEC_type body,
             environment,
             frame,
             operands;
    SURVIVE_DEFAULT_C(Value);
    binding_thread = (vRG_type)Thread_Peek();
    procedure = binding_thread->prc;
    frame     = binding_thread->frm;
    operands  = binding_thread->opd;
    position  = binding_thread->pos;
    index = get_NBR(position);
    frame[index++] = Value;
    raw_operand_count = size_VEC(operands);
    if (index > raw_operand_count)
      { body        = procedure->bod;
        environment = procedure->env;
        frame[index] = NULL_PAIR;
        return evaluate_body_poke_M(body,
                                    environment,
                                    frame); }
    param_count = procedure->par;
    raw_param_count = get_NBR(param_count);
    if (index > raw_param_count)
    binding_thread = (vRG_type)Thread_Patch(Continue_vararg);
    position = make_NBR(index);
    binding_thread->lst = NULL_PAIR;
    binding_thread->pos = position;
    expression = operands[index];
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_vararg_bindings_C(PRZ_type Procedure)
  { aPL_type application_thread;
    vRG_type binding_thread;
    EXP_type expression;
    NBR_type c_function,
             frame_size,
             param_count;
    UNS_type raw_frame_size,
             raw_operand_count,
             raw_param_count;
    VEC_type body,
             environment,
             frame,
             operands;
    application_thread = (aPL_type)Thread_Peek();
    operands = application_thread->opd;
    param_count = Procedure->par;
    raw_param_count = get_NBR(param_count);
    raw_operand_count = size_VEC(operands);
    if (raw_param_count > raw_operand_count)
      return error_procedure_vararg(TFO_message,
                                    Procedure);
    frame_size = Procedure->siz;
    raw_frame_size = get_NBR(frame_size);
    SURVIVE_DYNAMIC_C(Procedure,
                      raw_frame_size);
    frame = Environment_Create_Frame_M(raw_frame_size);
    if (raw_operand_count == 0)
      { body        = Procedure->bod;
        environment = Procedure->env;
        frame[1] = NULL_PAIR;
        return evaluate_body_poke_M(body,
                                    environment,
                                    frame); }
    if (raw_param_count == 0)
      c_function = Continue_vararg;
    else
      c_function = Continue_vararg_bindings;
    application_thread = (aPL_type)Thread_Peek();
    operands = application_thread->opd;
    binding_thread = (vRG_type)Thread_Poke_M(c_function,
                                             vRG_size);
    binding_thread->prc = Procedure;
    binding_thread->opd = operands;
    binding_thread->lst = NULL_PAIR;
    binding_thread->pos = Grammar_One;
    binding_thread->frm = frame;
    expression = operands[1];
    return evaluate_expression_C(expression); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_vararg_bindings_DC(VEC_type Frame)
  { EXP_type value;
    NBR_type frame_size,
             param_count;
    PAI_type arguments;
    PRZ_type procedure;
    UNS_type raw_frame_size,
             raw_param_count,
             position;
    VEC_type body,
             environment,
             frame;
    procedure = Frame[1];
    frame_size = procedure->siz;
    raw_frame_size = get_NBR(frame_size);
    SURVIVE_DYNAMIC_C(Frame,
                      raw_frame_size);
    procedure = Frame[1];
    arguments = Frame[2];
    param_count = procedure->par;
    body        = procedure->bod;
    environment = procedure->env;
    raw_param_count = get_NBR(param_count);
    if (raw_frame_size == 0)
      return error_procedure_vararg(IFS_message,
                                    procedure);
    frame = Environment_Create_Frame_M(raw_frame_size);
    for (position = 1;
         position <= raw_param_count;
         position++)
      { if (!is_PAI(arguments))
          return error_procedure_vararg(TFO_message,
                                        procedure);
        value     = arguments->car;
        arguments = arguments->cdr;
        frame[position] = value; }
    frame[position] = arguments;
    return evaluate_body_push_M(body,
                                environment,
                                frame); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_vararg_bindings(VOI_type)
  { Continue_vararg          = Thread_Register(continue_vararg_C);
    Continue_vararg_bindings = Thread_Register(continue_vararg_bindings_C); }

/*----------------------------------------------------------------------------*/
/*---------------------------------- while -----------------------------------*/
/*----------------------------------------------------------------------------*/

static NBR_type Continue_while_body,
                Continue_while_predicate;

BEGIN_THREAD_FRAME(wHI);
 THREAD_FRAME_SLOT(EXP, prd);
 THREAD_FRAME_SLOT(NBR, psz);
 THREAD_FRAME_SLOT(VEC, bod);
 THREAD_FRAME_SLOT(NBR, bsz);
 THREAD_FRAME_SLOT(EXP, res);
END_THREAD_FRAME(wHI);

static EXP_type continue_while_predicate_C(EXP_type Boolean)
  { wHI_type while_thread;
    EXP_type result;
    NBR_type body_size;
    UNS_type raw_body_size;
    VEC_type body;
    while_thread = (wHI_type)Thread_Peek();
    if (is_FLS(Boolean))
      { result = while_thread->res;
        Thread_Zap();
        return result; }
    while_thread = (wHI_type)Thread_Patch(Continue_while_body);
    body      = while_thread->bod;
    body_size = while_thread->bsz;
    raw_body_size = get_NBR(body_size);
    return evaluate_inline_sequence_C(body,
                                      raw_body_size); }

static EXP_type continue_while_body_C(EXP_type Value)
  { wHI_type while_thread;
    EXP_type predicate;
    NBR_type predicate_size;
    UNS_type raw_predicate_size;
    while_thread = (wHI_type)Thread_Peek();
    predicate      = while_thread->prd;
    predicate_size = while_thread->psz;
    while_thread = (wHI_type)Thread_Patch(Continue_while_predicate);
    while_thread->res = Value;
    raw_predicate_size = get_NBR(predicate_size);
    return evaluate_inline_push_C(predicate,
                                  raw_predicate_size); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_while_C(WHI_type While)
  { wHI_type while_thread;
    EXP_type predicate;
    NBR_type body_size,
             predicate_size;
    UNS_type raw_predicate_size;
    VEC_type body;
    SURVIVE_DEFAULT_C(While);
    predicate      = While->prd;
    predicate_size = While->psz;
    body           = While->bod;
    body_size      = While->bsz;
    while_thread = (wHI_type)Thread_Push_M(Continue_while_predicate,
                                           wHI_size);
    while_thread->prd = predicate;
    while_thread->psz = predicate_size;
    while_thread->bod = body;
    while_thread->bsz = body_size;
    while_thread->res = Grammar_Unspecified;
    raw_predicate_size = get_NBR(predicate_size);
    SURVIVE_DEFAULT_C(predicate);
    return evaluate_inline_push_C(predicate,
                                  raw_predicate_size); }

/*----------------------------------------------------------------------------*/

static VOI_type initialize_while(VOI_type)
  { Continue_while_body      = Thread_Register(continue_while_body_C);
    Continue_while_predicate = Thread_Register(continue_while_predicate_C); }

/*----------------------------------------------------------------------------*/
/*-------------------------------- expression --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_expression_C(EXP_type Expression)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case APL_tag:
          return evaluate_application_C(Expression);
        case BEG_tag:
          return evaluate_begin_C(Expression);
        case DFF_tag:
          return evaluate_define_function_C(Expression);
        case DFV_tag:
          return evaluate_define_variable_C(Expression);
        case DFZ_tag:
          return evaluate_define_function_vararg_C(Expression);
        case GLB_tag:
          return evaluate_global_variable(Expression);
        case IFF_tag:
          return evaluate_if_double_C(Expression);
        case IFZ_tag:
          return evaluate_if_single_C(Expression);
        case LCL_tag:
          return evaluate_local_variable(Expression);
        case LMB_tag:
          return evaluate_lambda_C(Expression);
        case LMZ_tag:
          return evaluate_lambda_vararg_C(Expression);
        case QUO_tag:
          return evaluate_quote_C(Expression);
        case STG_tag:
          return evaluate_set_global_C(Expression);
        case STL_tag:
          return evaluate_set_local_C(Expression);
        case WHI_tag:
          return evaluate_while_C(Expression);
        case CHA_tag:
        case FLS_tag:
        case NBR_tag:
        case NUL_tag:
        case PAI_tag:
        case REA_tag:
        case STR_tag:
        case SYM_tag:
        case TRU_tag:
        case VEC_tag:
          return evaluate_literal(Expression);
        default:
          return error_tag(IXT_message,
                           tag); }}

/*----------------------------- public functions -----------------------------*/

EXP_type Evaluate_C(EXP_type Expression)
  { EXP_type value;
    value = evaluate_expression_C(Expression);
    return value; }

EXP_type Evaluate_Apply_C(VEC_type Frame)
  { return evaluate_application_DC(Frame); }

/*------------------------------ initialization ------------------------------*/

VOI_type Evaluate_Initialize(VOI_type)
  { initialize_application();
    initialize_bindings();
    initialize_body();
    initialize_define();
    initialize_if();
    initialize_native_call();
    initialize_sequence();
    initialize_set_global();
    initialize_set_local();
    initialize_vararg_bindings();
    initialize_while(); }
