                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 9: garbage collection   */
                     /*-----------------------------------*/
                   /*               grammar             */
                     /*-----------------------------------*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SlipGrammar.h"

#include "SlipMemory.h"

/*------------------------------ private macros ------------------------------*/

#define CHUNK_SIZE(TYP)                                                        \
  (sizeof(TYP) / sizeof(CEL_type) - 1)

#define CHUNK_WITH_OFFSET(TYP, SIZ)                                            \
  (TYP##_type)Memory_Make_Chunk_M(TYP##_tag, TYP##_size + SIZ)

#define CHUNK(TYP)                                                             \
  CHUNK_WITH_OFFSET(TYP, 0)

/*----------------------------- private constants ----------------------------*/

enum { Immediate_null_value        = Memory_Void_Value + 0*Memory_Cell_Bias,
       Immediate_true_value        = Memory_Void_Value + 1*Memory_Cell_Bias,
       Immediate_false_value       = Memory_Void_Value + 2*Memory_Cell_Bias,
       Immediate_unspecified_value = Memory_Void_Value + 3*Memory_Cell_Bias,
       Immediate_boundary          = Memory_Void_Value + 3*Memory_Cell_Bias };

enum {    Default_claim_margin = 32,
       Minimum_number_of_cells = 2048,
                     Root_size = 22 };

static const TXT_type IMM_message = "insufficient memory";
static const TXT_type NAL_message = "not a list";
static const TXT_type NOF_message = "number overflow";

/*----------------------------- private variables ----------------------------*/

static EXP_type Survivor_P;

/*----------------------------------------------------------------------------*/

static VEC_type Root;
static UNS_type Root_counter;

static REF_type Force_collect[Root_size];

/*----------------------------- public variables -----------------------------*/

VEC_type Grammar_Empty_Vector_P;
FLS_type Grammar_False;
NUL_type Grammar_Null;
PAI_type Grammar_Null_Pair;
NBR_type Grammar_One;
TRU_type Grammar_True;
USP_type Grammar_Unspecified;
NBR_type Grammar_Zero;

/*----------------------------- private functions ----------------------------*/

static VOI_type error(TXT_type Message)
  { Main_Error(Message); }

static VOI_type error_text(TXT_type Message,
                           TXT_type Text )
  { sprintf(Main_Buffer, "%s: %s", Message, Text);
    error(Main_Buffer); }

static TAG_type tag_of(EXP_type Expression)
  { UNS_type immediate;
    if (Memory_Is_Immediate(Expression))
      return NBR_tag;
    immediate = (UNS_type)Expression;
    if (immediate <= Immediate_boundary)
      switch (immediate)
        { case Immediate_null_value:
            return NUL_tag;
          case Immediate_true_value:
            return TRU_tag;
          case Immediate_false_value:
            return FLS_tag;
          case Immediate_unspecified_value:
            return USP_tag; }
    return (TAG_type)Memory_Get_Tag((PTR_type)Expression); }

/*----------------------------------------------------------------------------*/

static BLN_type member_of(EXP_type Expression,
                          TAG_type Tag)
  { TAG_type tag;
    tag = tag_of(Expression);
    return (tag == Tag); }

static VOI_type reclaim_C(UNS_type Margin)
  { EXP_type expression;
    REF_type reference;
    UNS_type index;
    UNS_type after;
    for (index = 0;
         index < Root_counter;
         index++)
      { reference = Force_collect[index];
        Root[index + 1] = *reference; }
    Main_GC();
    Root = (VEC_type)Memory_Collect_C((PTR_type)Root);
    after = Memory_Available();
    for (index = 0;
         index < Root_counter;
         index++)
      { reference = Force_collect[index];
        expression = Root[index + 1];
        *reference = expression; }
    if (after < Margin)
      error(IMM_message); }

static UNS_type string_to_expression_size(TXT_type Raw_string)
  { return strlen(Raw_string) / sizeof(CEL_type) + 1; }

/*----------------------------- hidden functions -----------------------------*/

VOI_type _Grammar_Register(REF_type Reference)
  { assert(Root_counter < Root_size);
    Force_collect[Root_counter++] = Reference; }

VOI_type _Grammar_Survive_Default_C(REF_type Expression)
  { if (Memory_Claim(Default_claim_margin))
      { Survivor_P = *Expression;
        reclaim_C(Default_claim_margin);
        *Expression = Survivor_P;
        Survivor_P = Grammar_Null; }}

VOI_type _Grammar_Survive_Dynamic_C(REF_type Expression,
                                    UNS_type Margin)
  { if (Memory_Claim(Margin + Default_claim_margin))
      { Survivor_P = *Expression;
        reclaim_C(Margin + Default_claim_margin);
        *Expression = Survivor_P;
        Survivor_P = Grammar_Null; }}

/*----------------------------- public functions -----------------------------*/

VOI_type Grammar_Default_Claim_C(VOI_type)
  { if (Memory_Claim(Default_claim_margin))
      reclaim_C(Default_claim_margin); }

VOI_type Grammar_Dynamic_Claim_C(UNS_type Margin)
  { if (Memory_Claim(Margin + Default_claim_margin))
      reclaim_C(Margin + Default_claim_margin); }

UNS_type Grammar_List_Length(EXP_type List_expression,
                             TXT_type Text)
  { PAI_type pair;
    UNS_type length;
    for (length = 0;
         is_PAI(List_expression);
         length += 1)
      { pair = List_expression;
        List_expression = pair->cdr; }
    if (!is_NUL(List_expression))
      error_text(NAL_message,
                 Text);
    return length; }

NBR_type Grammar_Number(DBL_type Raw_number)
  { NBR_type number;
    if ((llabs(Raw_number) > Memory_Immediate_Maximum))
      error(NOF_message);
    number = (NBR_type)make_NBR((SGN_type)Raw_number);
    return number; }

VOI_type Grammar_Reclaim_C(VOI_type)
  { reclaim_C(0); }

PAI_type Grammar_Reverse_List(PAI_type List_expression,
                              TXT_type Text)
  { PAI_type next,
             pair,
             previous;
    previous = Grammar_Null_Pair;
    for (pair = List_expression;
         is_PAI(pair);
         pair = next)
      { next = pair->cdr;
        pair->cdr = previous;
        previous = pair; }
    if (!is_NUL(pair))
      error_text(NAL_message,
                 Text);
    return previous; }

TAG_type Grammar_Tag(EXP_type Expression)
  { return tag_of(Expression); }

/*------------------------------ APL declarations ----------------------------*/

static const UNS_type APL_size = CHUNK_SIZE(APL);

BLN_type is_APL(EXP_type Expression)
  { return member_of(Expression,
                     APL_tag); }

APL_type make_APL_M(EXP_type Expression,
                    VEC_type Operands)
  { APL_type application = CHUNK(APL);
    application->exp = Expression;
    application->opr = Operands;
    return application; }

/*------------------------------ BEG declarations ----------------------------*/

static const UNS_type BEG_size = CHUNK_SIZE(BEG);

BLN_type is_BEG(EXP_type Expression)
  { return member_of(Expression,
                     BEG_tag); }

BEG_type make_BEG_M(VEC_type Sequence)
  { BEG_type begin = CHUNK(BEG);
    begin->seq = Sequence;
    return begin; }

/*------------------------------ CHA declarations ----------------------------*/

static const UNS_type CHA_size = CHUNK_SIZE(CHA);

BLN_type is_CHA(EXP_type Expression)
  { return member_of(Expression,
                     CHA_tag); }

CHA_type make_CHA_M(CHR_type Raw_character)
  { CHA_type character = CHUNK(CHA);
    character->chr = Raw_character;
    return character; }

/*------------------------------ DFF declarations ----------------------------*/

static const UNS_type DFF_size = CHUNK_SIZE(DFF);

BLN_type is_DFF(EXP_type Expression)
  { return member_of(Expression,
                    DFF_tag); }

DFF_type make_DFF_M(SYM_type Name,
                    NBR_type Offset,
                    NBR_type Param_count,
                    NBR_type Frame_size,
                    VEC_type Body)
  { DFF_type definition = CHUNK(DFF);
    definition->nam = Name;
    definition->ofs = Offset;
    definition->par = Param_count;
    definition->siz = Frame_size;
    definition->bod = Body;
    return definition; }

/*------------------------------ DFV declarations ----------------------------*/

static const UNS_type DFV_size = CHUNK_SIZE(DFV);

BLN_type is_DFV(EXP_type Expression)
  { return member_of(Expression,
                     DFV_tag); }

DFV_type make_DFV_M(NBR_type Offset,
                    EXP_type Expression)
  { DFV_type definition = CHUNK(DFV);
    definition->ofs = Offset;
    definition->exp = Expression;
    return definition; }

/*------------------------------ DFZ declarations ----------------------------*/

static const UNS_type DFZ_size = CHUNK_SIZE(DFZ);

BLN_type is_DFZ(EXP_type Expression)
  { return member_of(Expression,
                     DFZ_tag); }

DFZ_type make_DFZ_M(SYM_type Name,
                    NBR_type Offset,
                    NBR_type Param_count,
                    NBR_type Frame_size,
                    VEC_type Body)
  { DFZ_type definition = CHUNK(DFZ);
    definition->nam = Name;
    definition->ofs = Offset;
    definition->par = Param_count;
    definition->siz = Frame_size;
    definition->bod = Body;
    return definition; }

/*------------------------------ FLS declarations ----------------------------*/

BLN_type is_FLS(EXP_type Expression)
  { return (Expression == (EXP_type)Immediate_false_value); }

FLS_type make_FLS(VOI_type)
  { return (FLS_type)Immediate_false_value; }


/*------------------------------ GLB declarations ----------------------------*/

const static UNS_type GLB_size = CHUNK_SIZE(GLB);

BLN_type is_GLB(EXP_type Expression)
  { return member_of(Expression,
                     GLB_tag); }

GLB_type make_GLB_M(NBR_type Scope,
                    NBR_type Offset)
  { GLB_type global = CHUNK(GLB);
    global->scp = Scope;
    global->ofs = Offset;
    return global; }

/*------------------------------ IFF declarations ----------------------------*/

static const UNS_type IFF_size = CHUNK_SIZE(IFF);

BLN_type is_IFF(EXP_type Expression)
  { return member_of(Expression,
                     IFF_tag); }

IFF_type make_IFF_M(EXP_type Predicate,
                    EXP_type Consequent,
                    NBR_type Consequent_size,
                    EXP_type Alternative,
                    NBR_type Alternative_size)
  { IFF_type if_ = CHUNK(IFF);
    if_->prd = Predicate;
    if_->cns = Consequent;
    if_->csz = Consequent_size;
    if_->alt = Alternative;
    if_->asz = Alternative_size;
    return if_; }

/*------------------------------ IFZ declarations ----------------------------*/

static const UNS_type IFZ_size = CHUNK_SIZE(IFZ);

BLN_type is_IFZ(EXP_type Expression)
  { return member_of(Expression,
                     IFZ_tag); }

IFZ_type make_IFZ_M(EXP_type Predicate,
                    EXP_type Consequent,
                    NBR_type Consequent_size)
  { IFZ_type if_ = CHUNK(IFZ);
    if_->prd = Predicate;
    if_->cns = Consequent;
    if_->csz = Consequent_size;
    return if_; }

/*------------------------------ LCL declarations ----------------------------*/

const static UNS_type LCL_size = CHUNK_SIZE(LCL);

BLN_type is_LCL(EXP_type Expression)
  { return member_of(Expression,
                     LCL_tag); }

LCL_type make_LCL_M(NBR_type Offset)
  { LCL_type local = CHUNK(LCL);
    local->ofs = Offset;
    return local; }

/*------------------------------ LMB declarations ----------------------------*/

static const UNS_type LMB_size = CHUNK_SIZE(LMB);

BLN_type is_LMB(EXP_type Expression)
  { return member_of(Expression,
                     LMB_tag); }

LMB_type make_LMB_M(NBR_type Param_count,
                    NBR_type Frame_size,
                    VEC_type Body)
  { LMB_type lambda = CHUNK(LMB);
    lambda->par = Param_count;
    lambda->siz = Frame_size;
    lambda->bod = Body;
    return lambda; }

/*------------------------------ LMZ declarations ----------------------------*/

static const UNS_type LMZ_size = CHUNK_SIZE(LMZ);

BLN_type is_LMZ(EXP_type Expression)
  { return member_of(Expression,
                     LMZ_tag); }

LMZ_type make_LMZ_M(NBR_type Param_count,
                    NBR_type Frame_size,
                    VEC_type Body)
  { LMZ_type lambda = CHUNK(LMZ);
    lambda->par = Param_count;
    lambda->siz = Frame_size;
    lambda->bod = Body;
    return lambda; }

/*------------------------------ NAT declarations ----------------------------*/

static const UNS_type NAT_size = CHUNK_SIZE(NAT);

BLN_type is_NAT(EXP_type Expression)
  { return member_of(Expression,
                     NAT_tag); }

NAT_type make_NAT_M(FUN_type Raw_function,
                    TXT_type Text)
  { NAT_type native;
    UNS_type size;
    size = string_to_expression_size(Text);
    native = CHUNK_WITH_OFFSET(NAT,
                               size);
    native->fun = Raw_function;
    strcpy(native->txt,
           Text);
    return native; }

/*------------------------------ NBR declarations ----------------------------*/

BLN_type is_NBR(EXP_type Expression)
  { return Memory_Is_Immediate(Expression); }

NBR_type make_NBR(SGN_type Raw_number)
  { NBR_type number;
    number = Memory_Make_Immediate(Raw_number);
    return number; }

SGN_type get_NBR(NBR_type Number)
  { SGN_type raw_number;
    raw_number = Memory_Get_Immediate(Number);
    return raw_number; }

/*------------------------------ NUL declarations ----------------------------*/

BLN_type is_NUL(EXP_type Expression)
  { return (Expression == Immediate_null_value); }

NUL_type make_NUL(VOI_type)
  { return (NUL_type)Immediate_null_value; }

/*------------------------------ PAI declarations ----------------------------*/

static const UNS_type PAI_size = CHUNK_SIZE(PAI);

BLN_type is_PAI(EXP_type Expression)
  { return member_of(Expression,
                     PAI_tag); }

PAI_type make_PAI_M(EXP_type Car,
                    EXP_type Cdr)
  { PAI_type pair = CHUNK(PAI);
    pair->car = Car;
    pair->cdr = Cdr;
    return pair; }

/*------------------------------ PRC declarations ----------------------------*/

static const UNS_type PRC_size = CHUNK_SIZE(PRC);

BLN_type is_PRC(EXP_type Expression)
  { return member_of(Expression,
                     PRC_tag); }

PRC_type make_PRC_M(SYM_type Name,
                    NBR_type Param_count,
                    NBR_type Frame_size,
                    VEC_type Body,
                    VEC_type Environment)
  { PRC_type procedure = CHUNK(PRC);
    procedure->nam = Name;
    procedure->par = Param_count;
    procedure->siz = Frame_size;
    procedure->bod = Body;
    procedure->env = Environment;
    return procedure; }

/*------------------------------ PRZ declarations ----------------------------*/

static const UNS_type PRZ_size = CHUNK_SIZE(PRZ);

BLN_type is_PRZ(EXP_type Expression)
  { return member_of(Expression,
                     PRZ_tag); }

PRZ_type make_PRZ_M(SYM_type Name,
                    NBR_type Param_count,
                    NBR_type Frame_size,
                    VEC_type Body,
                    VEC_type Environment)
  { PRZ_type procedure = CHUNK(PRZ);
    procedure->nam = Name;
    procedure->par = Param_count;
    procedure->siz = Frame_size;
    procedure->bod = Body;
    procedure->env = Environment;
    return procedure; }

/*------------------------------ QUO declarations ----------------------------*/

static const UNS_type QUO_size = CHUNK_SIZE(QUO);

BLN_type is_QUO(EXP_type Expression)
  { return member_of(Expression,
                     QUO_tag); }

QUO_type make_QUO_M(EXP_type Expression)
  { QUO_type quote = CHUNK(QUO);
    quote->exp = Expression;
    return quote; }

/*------------------------------ REA declarations ----------------------------*/

static const UNS_type REA_size = CHUNK_SIZE(REA);

BLN_type is_REA(EXP_type Expression)
  { return member_of(Expression,
                     REA_tag); }

REA_type make_REA_M(FLO_type Raw_real)
  { REA_type real = CHUNK(REA);
    real->flo = Raw_real;
    return real; }

/*------------------------------ STG declarations ----------------------------*/

static const UNS_type STG_size = CHUNK_SIZE(STG);

BLN_type is_STG(EXP_type Expression)
  { return member_of(Expression,
                     STG_tag); }

STG_type make_STG_M(NBR_type Scope,
                    NBR_type Offset,
                    EXP_type Expression)
  { STG_type set = CHUNK(STG);
    set->scp = Scope;
    set->ofs = Offset;
    set->exp = Expression;
    return set; }

/*------------------------------ STL declarations ----------------------------*/

static const UNS_type STL_size = CHUNK_SIZE(STL);

BLN_type is_STL(EXP_type Expression)
  { return member_of(Expression,
                     STL_tag); }

STL_type make_STL_M(NBR_type Offset,
                    EXP_type Expression)
  { STL_type set = CHUNK(STL);
    set->ofs = Offset;
    set->exp = Expression;
    return set; }

/*------------------------------ STR declarations ----------------------------*/

static const UNS_type STR_size = CHUNK_SIZE(STR);

BLN_type is_STR(EXP_type Expression)
  { return member_of(Expression,
                     STR_tag); }

STR_type make_STR_M(TXT_type Raw_string)
  { STR_type string;
    UNS_type size;
    size = string_to_expression_size(Raw_string);
    string = CHUNK_WITH_OFFSET(STR,
                               size);
    strcpy(string->txt,
           Raw_string);
    return string; }

UNS_type size_STR(TXT_type Raw_string)
  { return (UNS_type)string_to_expression_size(Raw_string); }

/*------------------------------ SYM declarations ----------------------------*/

static const UNS_type SYM_size = CHUNK_SIZE(SYM);

BLN_type is_SYM(EXP_type Expression)
  { return member_of(Expression,
                     SYM_tag); }

SYM_type make_SYM_M(TXT_type Raw_string)
  { SYM_type symbol;
    UNS_type size;
    size = string_to_expression_size(Raw_string);
    symbol = CHUNK_WITH_OFFSET(SYM,
                               size);
    strcpy(symbol->txt,
           Raw_string);
    return symbol; }

UNS_type size_SYM(TXT_type Raw_string)
  { return (UNS_type)string_to_expression_size(Raw_string); }

/*------------------------------ TRU declarations ----------------------------*/

BLN_type is_TRU(EXP_type Expression)
  { return (Expression == (EXP_type)Immediate_true_value); }

TRU_type make_TRU(VOI_type)
  { return (TRU_type)Immediate_true_value; }

/*------------------------------ USP declarations ----------------------------*/

BLN_type is_USP(EXP_type Expression)
  { return (Expression == (EXP_type)Immediate_unspecified_value); }

USP_type make_USP(VOI_type)
  { return (USP_type)Immediate_unspecified_value; }

/*------------------------------ VEC declarations ----------------------------*/

static const UNS_type VEC_size = 0;

BLN_type is_VEC(EXP_type Expression)
  { return member_of(Expression,
                     VEC_tag); }

VEC_type make_VEC_M(UNS_type Size)
  { UNS_type index;
    VEC_type vector = (VEC_type)CHUNK_WITH_OFFSET(VEC,
                                                  Size);
    for (index = 1;
         index <= Size;
         index++)
      vector[index] = (EXP_type)Immediate_unspecified_value;
    return vector; }

UNS_type size_VEC(VEC_type Vector)
  { return (UNS_type)Memory_Get_Size((PTR_type)Vector); }

/*------------------------------ WHI declarations ----------------------------*/

static const UNS_type WHI_size = CHUNK_SIZE(WHI);

BLN_type is_WHI(EXP_type Expression)
  { return member_of(Expression,
                     WHI_tag); }

WHI_type make_WHI_M(EXP_type Predicate,
                    NBR_type Predicate_size,
                    VEC_type Body,
                    NBR_type Body_size)
  { WHI_type while_ = CHUNK(WHI);
    while_->prd = Predicate;
    while_->psz = Predicate_size;
    while_->bod = Body;
    while_->bsz = Body_size;
    return while_; }

/*------------------------------ initialization ------------------------------*/

VOI_type Grammar_Initialize_M(ADR_type Memory,
                              UNS_type Size)
  { UNS_type total_number_of_cells;
    total_number_of_cells = Size / Memory_Cell_Size;
    if (total_number_of_cells < Minimum_number_of_cells)
      Main_Fatal_Error(IMM_message);
    Memory_Initialize(Memory,
                      total_number_of_cells);
    Grammar_Empty_Vector_P     = make_VEC_M(0);
    Grammar_False              = make_FLS();
    Grammar_Null               = make_NUL();
    Grammar_One                = make_NBR(1);
    Grammar_True               = make_TRU();
    Grammar_Unspecified        = make_USP();
    Grammar_Zero               = make_NBR(0);
    Root = make_VEC_M(Root_size);
    Root_counter = 0;
    Survivor_P = Grammar_Null;
    REGISTER(Grammar_Empty_Vector_P);
    REGISTER(Survivor_P); }
