                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 9: garbage collection   */
                     /*-----------------------------------*/
                     /*               grammar             */
                     /*-----------------------------------*/

#include "SlipMain.h"

/*------------------------------ public macros -------------------------------*/

#define REGISTER(VAR)                                                          \
  _Grammar_Register((REF_type)&VAR)

#define SURVIVE_DEFAULT_C(SRV)                                                 \
  _Grammar_Survive_Default_C((REF_type)&SRV)

#define SURVIVE_DYNAMIC_C(SRV, SIZ)                                            \
  _Grammar_Survive_Dynamic_C((REF_type)&SRV, SIZ)

/*----------------------------- public constants -----------------------------*/

enum { Grammar_Global_Frame_Size = 256 };

/*------------------------------- public types -------------------------------*/

typedef enum { APL_tag = 0x00<<1 | 0x0,
               BEG_tag = 0x01<<1 | 0x0,
               DFF_tag = 0x02<<1 | 0x0,
               DFV_tag = 0x03<<1 | 0x0,
               DFZ_tag = 0x04<<1 | 0x0,
               GLB_tag = 0x05<<1 | 0x0,
               IFF_tag = 0x06<<1 | 0x0,
               IFZ_tag = 0x07<<1 | 0x0,
               LCL_tag = 0x08<<1 | 0x0,
               LMB_tag = 0x09<<1 | 0x0,
               LMZ_tag = 0x0A<<1 | 0x0,
               PAI_tag = 0x0B<<1 | 0x0,
               PRC_tag = 0x0C<<1 | 0x0,
               PRZ_tag = 0x0D<<1 | 0x0,
               QUO_tag = 0x0E<<1 | 0x0,
               STG_tag = 0x0F<<1 | 0x0,
               STL_tag = 0x10<<1 | 0x0,
               VEC_tag = 0x11<<1 | 0x0,
               WHI_tag = 0x12<<1 | 0x0,

               CHA_tag = 0x00<<1 | 0x1,
               FLS_tag = 0x01<<1 | 0x1,
               NAT_tag = 0x02<<1 | 0x1,
               NUL_tag = 0x03<<1 | 0x1,
               NBR_tag = 0x04<<1 | 0x1,
               REA_tag = 0x05<<1 | 0x1,
               STR_tag = 0x06<<1 | 0x1,
               SYM_tag = 0x07<<1 | 0x1,
               TRU_tag = 0x08<<1 | 0x1,
               USP_tag = 0x09<<1 | 0x1 } TAG_type;

typedef struct APL * APL_type;
typedef struct BEG * BEG_type;
typedef struct CHA * CHA_type;
typedef struct DFF * DFF_type;
typedef struct DFV * DFV_type;
typedef struct DFZ * DFZ_type;
typedef struct GLB * GLB_type;
typedef struct IFF * IFF_type;
typedef struct IFZ * IFZ_type;
typedef struct LCL * LCL_type;
typedef struct LMB * LMB_type;
typedef struct LMZ * LMZ_type;
typedef struct NAT * NAT_type;
typedef struct PAI * PAI_type;
typedef struct QUO * QUO_type;
typedef struct PRC * PRC_type;
typedef struct PRZ * PRZ_type;
typedef struct REA * REA_type;
typedef struct STG * STG_type;
typedef struct STL * STL_type;
typedef struct STR * STR_type;
typedef struct SYM * SYM_type;
typedef struct WHI * WHI_type;

typedef VOI_type * EXP_type;

typedef EXP_type   FLS_type;
typedef EXP_type   NBR_type;
typedef EXP_type   NUL_type;
typedef EXP_type * REF_type;
typedef EXP_type   TRU_type;
typedef EXP_type   USP_type;
typedef EXP_type * VEC_type;

typedef UNS_type * URF_type;

typedef EXP_type (* CCC_type) (EXP_type);
typedef EXP_type (* FUN_type) (VEC_type);

/*----------------------------- public variables -----------------------------*/

extern VEC_type Grammar_Empty_Vector_P;
extern FLS_type Grammar_False;
extern NUL_type Grammar_Null;
extern NBR_type Grammar_One;
extern TRU_type Grammar_True;
extern USP_type Grammar_Unspecified;
extern NBR_type Grammar_Zero;

/*----------------------------- hidden prototypes ----------------------------*/

VOI_type          _Grammar_Register(REF_type);
VOI_type _Grammar_Survive_Default_C(REF_type);
VOI_type _Grammar_Survive_Dynamic_C(REF_type,
                                    UNS_type);

/*----------------------------- public prototypes ----------------------------*/

VOI_type    Grammar_Default_Claim_C(VOI_type);
VOI_type    Grammar_Dynamic_Claim_C(UNS_type);
UNS_type        Grammar_List_Length(EXP_type,
                                    TXT_type);
NBR_type             Grammar_Number(DBL_type);
VOI_type          Grammar_Reclaim_C(VOI_type);
PAI_type       Grammar_Reverse_List(PAI_type,
                                    TXT_type);
TAG_type                Grammar_Tag(EXP_type);

/*----------------------------------- APL ------------------------------------*/

typedef struct APL { CEL_type hdr;
                     EXP_type exp;
                     VEC_type opr; } APL;

BLN_type                     is_APL(EXP_type);
APL_type                 make_APL_M(EXP_type,
                                    VEC_type);

/*----------------------------------- BEG ------------------------------------*/

typedef struct BEG { CEL_type hdr;
                     VEC_type seq; } BEG;

BLN_type                     is_BEG(EXP_type);
BEG_type                 make_BEG_M(VEC_type);

/*----------------------------------- CHA ------------------------------------*/

typedef struct CHA { CEL_type hdr;
                     CHR_type chr; } CHA;

BLN_type                     is_CHA(EXP_type);
CHA_type                 make_CHA_M(CHR_type);

/*----------------------------------- DFF ------------------------------------*/

typedef struct DFF { CEL_type hdr;
                     SYM_type nam;
                     NBR_type ofs;
                     NBR_type par;
                     NBR_type siz;
                     VEC_type bod; } DFF;

BLN_type                     is_DFF(EXP_type);
DFF_type                 make_DFF_M(SYM_type,
                                    NBR_type,
                                    NBR_type,
                                    NBR_type,
                                    VEC_type);

/*----------------------------------- DFV ------------------------------------*/

typedef struct DFV { CEL_type hdr;
                     NBR_type ofs;
                     EXP_type exp; } DFV;

BLN_type                     is_DFV(EXP_type);
DFV_type                 make_DFV_M(NBR_type,
                                    EXP_type);

/*----------------------------------- DFZ ------------------------------------*/

typedef struct DFZ { CEL_type hdr;
                     SYM_type nam;
                     NBR_type ofs;
                     NBR_type par;
                     NBR_type siz;
                     VEC_type bod; } DFZ;

BLN_type                     is_DFZ(EXP_type);
DFZ_type                 make_DFZ_M(SYM_type,
                                    NBR_type,
                                    NBR_type,
                                    NBR_type,
                                    VEC_type);

/*----------------------------------- FLS ------------------------------------*/

BLN_type                     is_FLS(EXP_type);
FLS_type                   make_FLS(VOI_type);

/*----------------------------------- GLB ------------------------------------*/

typedef struct GLB { CEL_type hdr;
                     NBR_type scp;
                     NBR_type ofs; } GLB;

BLN_type                     is_GLB(EXP_type);
GLB_type                 make_GLB_M(NBR_type,
                                    NBR_type);

/*----------------------------------- IFF ------------------------------------*/

typedef struct IFF { CEL_type hdr;
                     EXP_type prd;
                     EXP_type cns;
                     NBR_type csz;
                     EXP_type alt;
                     NBR_type asz; } IFF;

BLN_type                     is_IFF(EXP_type);
IFF_type                 make_IFF_M(EXP_type,
                                    EXP_type,
                                    NBR_type,
                                    EXP_type,
                                    NBR_type);

/*----------------------------------- IFZ ------------------------------------*/

typedef struct IFZ { CEL_type hdr;
                     EXP_type prd;
                     EXP_type cns;
                     NBR_type csz; } IFZ;

BLN_type                     is_IFZ(EXP_type);
IFZ_type                 make_IFZ_M(EXP_type,
                                    EXP_type,
                                    NBR_type);

/*----------------------------------- LCL ------------------------------------*/

typedef struct LCL { CEL_type hdr;
                     NBR_type ofs; } LCL;

BLN_type                     is_LCL(EXP_type);
LCL_type                 make_LCL_M(NBR_type);

/*----------------------------------- LMB ------------------------------------*/

typedef struct LMB { CEL_type hdr;
                     NBR_type par;
                     NBR_type siz;
                     VEC_type bod; } LMB;

BLN_type                     is_LMB(EXP_type);
LMB_type                 make_LMB_M(NBR_type,
                                    NBR_type,
                                    VEC_type);

/*----------------------------------- LMZ ------------------------------------*/

typedef struct LMZ { CEL_type hdr;
                     NBR_type par;
                     NBR_type siz;
                     VEC_type bod; } LMZ;

BLN_type                     is_LMZ(EXP_type);
LMZ_type                 make_LMZ_M(NBR_type,
                                    NBR_type,
                                    VEC_type);

/*----------------------------------- NAT ------------------------------------*/

typedef struct NAT { CEL_type hdr;
                     FUN_type fun;
                     CHR_type txt[]; } NAT;

BLN_type                     is_NAT(EXP_type);
NAT_type                 make_NAT_M(FUN_type,
                                    TXT_type);

/*----------------------------------- NBR ------------------------------------*/

BLN_type                     is_NBR(EXP_type);
NBR_type                   make_NBR(SGN_type);
SGN_type                    get_NBR(NBR_type);

/*----------------------------------- NUL ------------------------------------*/

BLN_type                     is_NUL(EXP_type);
NUL_type                   make_NUL(VOI_type);

/*----------------------------------- PAI ------------------------------------*/

typedef struct PAI { CEL_type hdr;
                     EXP_type car;
                     EXP_type cdr; } PAI;

BLN_type                     is_PAI(EXP_type);
PAI_type                 make_PAI_M(EXP_type,
                                    EXP_type);

/*----------------------------------- PRC ------------------------------------*/

typedef struct PRC { CEL_type hdr;
                     SYM_type nam;
                     NBR_type par;
                     NBR_type siz;
                     VEC_type bod;
                     VEC_type env; } PRC;

BLN_type                     is_PRC(EXP_type);
PRC_type                 make_PRC_M(SYM_type,
                                    NBR_type,
                                    NBR_type,
                                    VEC_type,
                                    VEC_type);

/*----------------------------------- PRZ ------------------------------------*/

typedef struct PRZ { CEL_type hdr;
                     SYM_type nam;
                     NBR_type par;
                     NBR_type siz;
                     VEC_type bod;
                     VEC_type env; } PRZ;

BLN_type                     is_PRZ(EXP_type);
PRZ_type                 make_PRZ_M(SYM_type,
                                    NBR_type,
                                    NBR_type,
                                    VEC_type,
                                    VEC_type);

/*----------------------------------- QUO ------------------------------------*/

typedef struct QUO { CEL_type hdr;
                     EXP_type exp; } QUO;

BLN_type                     is_QUO(EXP_type);
QUO_type                 make_QUO_M(EXP_type);

/*----------------------------------- REA ------------------------------------*/

typedef struct REA { CEL_type hdr;
                     FLO_type flo; } REA;

BLN_type                     is_REA(EXP_type);
REA_type                 make_REA_M(FLO_type);

/*----------------------------------- STG ------------------------------------*/

typedef struct STG { CEL_type hdr;
                     NBR_type scp;
                     NBR_type ofs;
                     EXP_type exp; } STG;

BLN_type                     is_STG(EXP_type);
STG_type                 make_STG_M(NBR_type,
                                    NBR_type,
                                    EXP_type);

/*----------------------------------- STL ------------------------------------*/

typedef struct STL { CEL_type hdr;
                     NBR_type ofs;
                     EXP_type exp; } STL;

BLN_type                     is_STL(EXP_type);
STL_type                 make_STL_M(NBR_type,
                                    EXP_type);

/*----------------------------------- STR ------------------------------------*/

typedef struct STR { CEL_type hdr;
                     CHR_type txt[]; } STR;

BLN_type                     is_STR(EXP_type);
STR_type                 make_STR_M(TXT_type);
UNS_type                   size_STR(TXT_type);

/*----------------------------------- SYM ------------------------------------*/

typedef struct SYM { CEL_type hdr;
                     CHR_type txt[]; } SYM;

BLN_type                     is_SYM(EXP_type);
SYM_type                 make_SYM_M(TXT_type);
UNS_type                   size_SYM(TXT_type);

/*----------------------------------- TRU ------------------------------------*/

BLN_type                     is_TRU(EXP_type);
TRU_type                   make_TRU(VOI_type);

/*----------------------------------- USP ------------------------------------*/

BLN_type                     is_USP(EXP_type);
USP_type                   make_USP(VOI_type);

/*----------------------------------- VEC ------------------------------------*/

BLN_type                     is_VEC(EXP_type);
VEC_type                 make_VEC_M(UNS_type);
UNS_type                   size_VEC(VEC_type);

/*----------------------------------- WHI ------------------------------------*/

typedef struct WHI { CEL_type hdr;
                     EXP_type prd;
                     NBR_type psz;
                     VEC_type bod;
                     NBR_type bsz; } WHI;

BLN_type                     is_WHI(EXP_type);
WHI_type                 make_WHI_M(EXP_type,
                                    NBR_type,
                                    VEC_type,
                                    NBR_type);

/*------------------------------ initialization ------------------------------*/

VOI_type       Grammar_Initialize_M(ADR_type,
                                    UNS_type);
