                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 9: garbage collection   */
                     /*-----------------------------------*/
                     /*               pool                */
                     /*-----------------------------------*/

#include <string.h>

#include "SlipGrammar.h"

#include "SlipPool.h"

/*------------------------------ private constants ---------------------------*/

enum {   Initial_pool_size = 64,
        Natives_table_size = 45,
       Pool_size_increment = 32 };

static const TXT_type NTO_message = "native table overflow";

/*----------------------------- private variables ----------------------------*/

static VEC_type Pool_vector_P;

static TXT_type Natives_table [Natives_table_size];
static UNS_type Pool_size;
static UNS_type Top_of_pool;

/*----------------------------- public variables -----------------------------*/

SYM_type Pool_Anonymous_P;
SYM_type Pool_Application_P;
SYM_type Pool_Begin_P;
SYM_type Pool_Define_P;
SYM_type Pool_If_P;
SYM_type Pool_Lambda_P;
SYM_type Pool_Load_P;
SYM_type Pool_Quote_P;
SYM_type Pool_Set_P;
SYM_type Pool_While_P;

/*----------------------------- public functions -----------------------------*/

SYM_type Pool_Enter_C(TXT_type Text)
  { SYM_type symbol;
    TXT_type text;
    UNS_type index,
             size;
    VEC_type vector;
    size = size_SYM(Text);
    for (index = 1;
         index < Top_of_pool;
         index++)
      { symbol = Pool_vector_P[index];
        text = symbol->txt;
        if (strcmp(Text,
                   text) == 0)
          return symbol; }
    if (Top_of_pool == Pool_size)
      { Pool_size += Pool_size_increment;
        Grammar_Dynamic_Claim_C(Pool_size);
        vector = Pool_vector_P;
        Pool_vector_P = make_VEC_M(Pool_size);
        for (index = 1;
             index < Top_of_pool;
             index++)
          Pool_vector_P[index] = vector[index]; }
    Grammar_Dynamic_Claim_C(size);
    symbol = make_SYM_M(Text);
    Pool_vector_P[Top_of_pool++] = symbol;
    return symbol; }

VOI_type Pool_Enter_Native_Name(TXT_type String,
                                UNS_type Offset)
  { if (Offset >= Natives_table_size)
      Main_Fatal_Error(NTO_message);
    Natives_table[Offset - 1] = String; }

TXT_type Pool_Get_Native_Name(UNS_type Offset)
  { if (Offset >= Natives_table_size)
      return 0;
    return Natives_table[Offset - 1]; }

/*------------------------------ initialization ------------------------------*/

VOI_type Pool_Initialize_M(VOI_type)
  { Pool_size = Initial_pool_size;
    Top_of_pool = 1;
    Pool_vector_P = make_VEC_M(Initial_pool_size);
    Pool_Anonymous_P   = Pool_Enter_C(Pool_Anonymous_Text);
    Pool_Application_P = Pool_Enter_C(Pool_Application_Text);
    Pool_Begin_P       = Pool_Enter_C(Pool_Begin_Text);
    Pool_Define_P      = Pool_Enter_C(Pool_Define_Text);
    Pool_If_P          = Pool_Enter_C(Pool_If_Text);
    Pool_Lambda_P      = Pool_Enter_C(Pool_Lambda_Text);
    Pool_Load_P        = Pool_Enter_C(Pool_Load_Text);
    Pool_Quote_P       = Pool_Enter_C(Pool_Quote_Text);
    Pool_Set_P         = Pool_Enter_C(Pool_Set_Text);
    Pool_While_P       = Pool_Enter_C(Pool_While_Text);
    REGISTER(Pool_Anonymous_P);
    REGISTER(Pool_Application_P);
    REGISTER(Pool_Begin_P);
    REGISTER(Pool_Define_P);
    REGISTER(Pool_If_P);
    REGISTER(Pool_Lambda_P);
    REGISTER(Pool_Load_P);
    REGISTER(Pool_Quote_P);
    REGISTER(Pool_Set_P);
    REGISTER(Pool_While_P);
    REGISTER(Pool_vector_P); }