                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 9: garbage collection   */
                     /*-----------------------------------*/
                     /*               print               */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipGrammar.h"

#include "SlipPool.h"

#include "SlipPrint.h"

/*------------------------------ private macros ------------------------------*/

#define FORMAT(FORMAT, VALUE)                                                  \
  { snprintf(Main_Buffer,                                                      \
             Main_Buffer_Size,                                                 \
             FORMAT,                                                           \
             VALUE);                                                           \
    Main_Print(Main_Buffer); }

/*----------------------------- private constants ----------------------------*/

static const TXT_type IXT_message = "invalid expression type";

static const TXT_type   Character_display_text = "%c";
static const TXT_type  Empty_list_display_text = "()";
static const TXT_type       False_display_text = "#f";
static const TXT_type     Leftpar_display_text = "(";
static const TXT_type     Leftvec_display_text = "[";
static const TXT_type      Native_display_text = "<native %s>";
static const TXT_type     Newline_display_text = "\n";
static const TXT_type      Number_display_text = "%ld";
static const TXT_type      Period_display_text = " . ";
static const TXT_type   Procedure_display_text = "<procedure %s>";
static const TXT_type        Real_display_text = "%#.10g";
static const TXT_type    Rightpar_display_text = ")";
static const TXT_type    Rightvec_display_text = "]";
static const TXT_type       Space_display_text = " ";
static const TXT_type        True_display_text = "#t";
static const TXT_type Unspecified_display_text = "";


static const TXT_type    Character_print_text = "#\\%c";
static const TXT_type   Empty_list_print_text = "()";
static const TXT_type        False_print_text = "#f";
static const TXT_type      Leftpar_print_text = "(";
static const TXT_type      Leftvec_print_text = "[";
static const TXT_type     Variable_print_text = "var_%lu";
static const TXT_type Variable_bis_print_text = "_%lu";
static const TXT_type       Native_print_text = "<native %s>";
static const TXT_type      Newline_print_text = "\n";
static const TXT_type       Number_print_text = "%ld";
static const TXT_type       Period_print_text = " . ";
static const TXT_type    Procedure_print_text = "<procedure %s>";
static const TXT_type        Quote_print_text = "'";
static const TXT_type         Real_print_text = "%#.10g";
static const TXT_type     Rightpar_print_text = ")";
static const TXT_type     Rightvec_print_text = "]";
static const TXT_type        Space_print_text = " ";
static const TXT_type       String_print_text = "\"%s\"";
static const TXT_type       Symbol_print_text = "%s";
static const TXT_type         True_print_text = "#t";
static const TXT_type  Unspecified_print_text = "";

/*----------------------------- private variables ----------------------------*/

static UNS_type Indentation;
static UNS_type       Scope;

/*---------------------------- private prototypes ----------------------------*/

static VOI_type display_expression(EXP_type);
static VOI_type   print_expression(EXP_type);

/*----------------------------- private functions ----------------------------*/

static VOI_type indent(VOI_type)
  { Indentation += 2; }

static VOI_type outdent(VOI_type)
  { Indentation -= 2; }

static VOI_type print_error_tag(TXT_type Message,
                                TAG_type Tag)
  { sprintf(Main_Buffer, "%s: %d", Message, Tag);
    Main_Error(Message);  }

/*---------------------------private display functions -----------------------*/

static VOI_type display_character(CHA_type Character)
  { FORMAT(Character_display_text,
           Character->chr); }

static VOI_type display_false(VOI_type)
  { Main_Print(False_display_text); }

static VOI_type display_native(NAT_type Native)
  { TXT_type text = Native->txt;
    FORMAT(Native_display_text,
           text); }

static VOI_type display_newline(VOI_type)
  { Main_Print(Newline_display_text); }

static VOI_type display_null(VOI_type)
  { Main_Print(Empty_list_display_text); }

static VOI_type display_number(NBR_type Number)
  { SGN_type raw_number;
    raw_number = get_NBR(Number);
    FORMAT(Number_display_text,
           raw_number); }

static VOI_type display_pair(PAI_type Pair)
  { EXP_type residue,
             value;
    PAI_type list;
    TAG_type tag;
    Main_Print(Leftpar_display_text);
    for (list = Pair;
         ;
         )
      { value   = list->car;
        residue = list->cdr;
        display_expression(value);
        tag = Grammar_Tag(residue);
        switch (tag)
          { case NUL_tag:
              Main_Print(Rightpar_display_text);
              return;
            case PAI_tag:
              Main_Print(Space_display_text);
              list = residue;
              break;
            default:
              Main_Print(Period_display_text);
              display_expression(residue);
              Main_Print(Rightpar_display_text);
              return; }}}

static VOI_type display_procedure(PRC_type Procedure)
  { SYM_type name;
    TXT_type text;
    name = Procedure->nam;
    text = name->txt;
    FORMAT(Procedure_display_text,
           text); }

static VOI_type display_procedure_vararg(PRZ_type Procedure)
  { SYM_type name;
    TXT_type text;
    name = Procedure->nam;
    text = name->txt;
    FORMAT(Procedure_display_text,
           text); }

static VOI_type display_real(REA_type Real)
  { FORMAT(Real_display_text,
           Real->flo); }

static VOI_type display_string(STR_type String)
  { TXT_type text;
    text = String->txt;
    Main_Print(text); }

static VOI_type display_symbol(SYM_type Symbol)
  { TXT_type text;
    text = Symbol->txt;
    Main_Print(text); }

static VOI_type display_true(VOI_type)
  { Main_Print(True_display_text); }

static VOI_type display_unspecified(VOI_type)
  { Main_Print(Unspecified_display_text); }

static VOI_type display_vector(VEC_type Vector)
  { EXP_type value;
    UNS_type index,
             size;
    size = size_VEC(Vector);
    Main_Print(Leftvec_display_text);
    for (index = 1;
         index <= size;
         index++)
      { value = Vector[index];
        display_expression(value);
        if (index < size)
          Main_Print(Space_display_text); }
    Main_Print(Rightvec_display_text); }

/*----------------------------------------------------------------------------*/

static VOI_type display_expression(EXP_type Value)
  { TAG_type tag;
    tag = Grammar_Tag(Value);
    switch (tag)
      { case CHA_tag:
          display_character(Value);
          return;
        case FLS_tag:
          display_false();
          return;
        case NAT_tag:
          display_native(Value);
          return;
        case NBR_tag:
          display_number(Value);
          return;
        case NUL_tag:
          display_null();
          return;
        case PAI_tag:
          display_pair(Value);
          return;
        case PRC_tag:
          display_procedure(Value);
          return;
        case PRZ_tag:
          display_procedure_vararg(Value);
          return;
        case REA_tag:
          display_real(Value);
          return;
        case STR_tag:
          display_string(Value);
          return;
        case SYM_tag:
          display_symbol(Value);
          return;
        case TRU_tag:
          display_true();
          return;
        case USP_tag:
          display_unspecified();
          return;
        case VEC_tag:
          display_vector(Value);
          return;
        default:
          print_error_tag(IXT_message,
                          tag); }}

/*--------------------------- private print functions ------------------------*/

static VOI_type print_newline(VOI_type)
  { UNS_type index;
    Main_Print(Newline_print_text);
    for (index = 0;
         index < Indentation;
         index++)
       Main_Print(Space_print_text); }

static VOI_type print_operands(VEC_type Operands)
  { EXP_type expression;
    UNS_type size,
             index;
 	  size = size_VEC(Operands);
    for (index = 1;
         index <= size;
         index++)
      { Main_Print(Space_print_text);
        expression = Operands[index];
        print_expression(expression); }}

static VOI_type print_sequence(VEC_type Sequence)
  { EXP_type expression;
    UNS_type index,
             size;
    size = size_VEC(Sequence);
    indent();
    for (index = 1;
         index <= size;
         index++)
      { print_newline();
        expression = Sequence[index];
        print_expression(expression); }
    outdent(); }

static VOI_type print_variable(UNS_type Scope,
                               UNS_type Offset)
  { TXT_type name;
    if (Scope == 1)
      { name = Pool_Get_Native_Name(Offset);
        if (name)
          { Main_Print(name);
            return; }}
    FORMAT(Variable_print_text,
           Scope);
    FORMAT(Variable_bis_print_text,
           Offset);  }

static VOI_type print_pattern(UNS_type Param_count)
  { UNS_type index;
    if (Param_count > 0)
      for (index = 1;
           index <= Param_count;
           index++)
        { if (index > 1)
            Main_Print(Space_print_text);
          print_variable(Scope,
                         index); }}

/*----------------------------------------------------------------------------*/

static VOI_type print_symbol(SYM_type);

static VOI_type print_application(APL_type Application)
  { EXP_type operator;
    VEC_type operands;
    operator = Application->exp;
    operands = Application->opr;
    Main_Print(Leftpar_print_text);
    print_expression(operator);
    print_operands(operands);
    Main_Print(Rightpar_print_text); }

static VOI_type print_begin(BEG_type Begin)
  { VEC_type sequence;
    sequence = Begin->seq;
    Main_Print(Leftpar_print_text);
    print_symbol(Pool_Begin_P);
    print_sequence(sequence);
    Main_Print(Rightpar_print_text); }

static VOI_type print_character(CHA_type Character)
  { CHR_type character;
    character = Character->chr;
    FORMAT(Character_print_text,
           character); }

static VOI_type print_define_function(DFF_type Define)
  { NBR_type offset,
             param_count;
    UNS_type raw_offset,
             raw_param_count;
    VEC_type body;
    offset      = Define->ofs;
    param_count = Define->par;
    body          = Define->bod;
    raw_offset = get_NBR(offset);
    raw_param_count = get_NBR(param_count);
    Main_Print(Leftpar_print_text);
    print_symbol(Pool_Define_P);
    Main_Print(Space_print_text);
    Main_Print(Leftpar_print_text);
    print_variable(Scope,
                   raw_offset);
    Scope++;
    Main_Print(Space_print_text);
    print_pattern(raw_param_count);
    Main_Print(Rightpar_print_text);
    print_sequence(body);
    Scope--;
    Main_Print(Rightpar_print_text);  }

static VOI_type print_define_function_vararg(DFZ_type Define)
  { NBR_type offset,
             param_count;
    UNS_type raw_offset,
             raw_param_count;
    VEC_type body;
    offset      = Define->ofs;
    param_count = Define->par;
    body          = Define->bod;
    raw_offset = get_NBR(offset);
    raw_param_count = get_NBR(param_count);
    Main_Print(Leftpar_print_text);
    print_symbol(Pool_Define_P);
    Main_Print(Space_print_text);
    Main_Print(Leftpar_print_text);
    print_variable(Scope,
                   raw_offset);
    Scope++;
    Main_Print(Space_print_text);
    print_pattern(raw_param_count);
    Main_Print(Period_print_text);
    print_variable(Scope,
                   raw_param_count + 1);
    Main_Print(Rightpar_print_text);
    print_sequence(body);
    Scope--;
    Main_Print(Rightpar_print_text);  }

static VOI_type print_define_variable(DFV_type Define)
  { EXP_type expression;
    NBR_type offset;
    UNS_type raw_offset;
    offset   = Define->ofs;
    raw_offset = get_NBR(offset);
    expression = Define->exp;
    Main_Print(Leftpar_print_text);
    print_symbol(Pool_Define_P);
    Main_Print(Space_print_text);
    print_variable(Scope,
                   raw_offset);
    Main_Print(Space_print_text);
    print_expression(expression);
    Main_Print(Rightpar_print_text); }

static VOI_type print_false(VOI_type)
  { Main_Print(False_print_text); }

static VOI_type print_global(GLB_type Global)
  { NBR_type offset,
             scope;
    UNS_type raw_offset,
             raw_scope;
    scope  = Global->scp;
    offset = Global->ofs;
    raw_scope = get_NBR(scope);
    raw_offset = get_NBR(offset);
    print_variable(raw_scope,
                   raw_offset); }

static VOI_type print_if_double(IFF_type If)
  { EXP_type alternative,
             consequent,
             predicate;
    predicate  = If->prd;
    consequent = If->cns;
    alternative = If->alt;
    Main_Print(Leftpar_print_text );
    print_symbol(Pool_If_P);
    Main_Print(Space_print_text);
    print_expression(predicate);
    indent();
    print_newline();
    Scope++;
    print_expression(consequent);
    Scope--;
    print_newline();
    Scope++;
    print_expression(alternative);
    Scope--;
    Main_Print(Rightpar_print_text);
    outdent(); }

static VOI_type print_if_single(IFZ_type If)
  { EXP_type consequent,
             predicate;
    predicate  = If->prd;
    consequent = If->cns;
    Main_Print(Leftpar_print_text );
    print_symbol(Pool_If_P);
    Main_Print(Space_print_text);
    print_expression(predicate);
    indent();
    print_newline();
    Scope++;
    print_expression(consequent);
    Scope--;
    Main_Print(Rightpar_print_text);
    outdent(); }

static VOI_type print_lambda(LMB_type Lambda)
  { NBR_type param_count;
    UNS_type raw_param_count;
    VEC_type body;
    param_count = Lambda->par;
    body          = Lambda->bod;
    raw_param_count = get_NBR(param_count);
    Main_Print(Leftpar_print_text);
    print_symbol(Pool_Lambda_P);
    Main_Print(Space_print_text);
    Main_Print(Leftpar_print_text);
    Scope++;
    print_pattern(raw_param_count);
    Main_Print(Rightpar_print_text);
    print_sequence(body);
    Scope--;
    Main_Print(Rightpar_print_text);  }

static VOI_type print_lambda_vararg(LMZ_type Lambda)
  { NBR_type param_count;
    UNS_type raw_param_count;
    VEC_type body;
    param_count = Lambda->par;
    body          = Lambda->bod;
    raw_param_count = get_NBR(param_count);
    Main_Print(Leftpar_print_text);
    print_symbol(Pool_Lambda_P);
    Main_Print(Space_print_text);
    Scope++;
    if (raw_param_count == 0)
      print_variable(Scope,
                     1);
    else
      { Main_Print(Leftpar_print_text);
        print_pattern(raw_param_count);
        Main_Print(Period_print_text);
        print_variable(Scope,
                       raw_param_count + 1);
        Main_Print(Rightpar_print_text); }
    print_sequence(body);
    Scope--;
    Main_Print(Rightpar_print_text);  }

static VOI_type print_local(LCL_type Local)
  { NBR_type offset;
    UNS_type raw_offset;
    offset = Local->ofs;
    raw_offset = get_NBR(offset);
    print_variable(Scope,
                   raw_offset); }

static VOI_type print_native(NAT_type Native)
  { TXT_type text = Native->txt;
    FORMAT(Native_print_text,
           text); }

static VOI_type print_null(VOI_type)
  { Main_Print(Empty_list_print_text); }

static VOI_type print_number(NBR_type Number)
  { SGN_type raw_number;
    raw_number = get_NBR(Number);
    FORMAT(Number_print_text,
           raw_number); }

static VOI_type print_pair(PAI_type Pair)
  { EXP_type residue,
             value;
    PAI_type list;
    TAG_type tag;
    Main_Print(Leftpar_print_text);
    for (list = Pair;
         ;
         )
      { value   = list->car;
        residue = list->cdr;
        print_expression(value);
        tag = Grammar_Tag(residue);
        switch (tag)
          { case NUL_tag:
              Main_Print(Rightpar_print_text);
              return;
            case PAI_tag:
              Main_Print(Space_print_text);
              list = residue;
              break;
            default:
              Main_Print(Period_print_text);
              print_expression(residue);
              Main_Print(Rightpar_print_text);
              return; }}}

static VOI_type print_procedure(PRC_type Procedure)
  { SYM_type name;
    TXT_type text;
    name = Procedure->nam;
    text = name->txt;
    FORMAT(Procedure_print_text,
           text); }

static VOI_type print_procedure_vararg(PRZ_type Procedure)
  { SYM_type name;
    TXT_type text;
    name = Procedure->nam;
    text = name->txt;
    FORMAT(Procedure_print_text,
           text); }

static VOI_type print_quote(QUO_type Quote)
  { EXP_type expression;
    expression = Quote->exp;
    Main_Print(Quote_print_text);
    print_expression(expression); }

static VOI_type print_real(REA_type Real)
  { FORMAT(Real_print_text,
           Real->flo); }

static VOI_type print_set_global(STG_type Set)
  { EXP_type expression;
    NBR_type offset,
             scope;
    UNS_type raw_offset,
             raw_scope;
    scope    = Set->scp;
    offset   = Set->ofs;
    expression = Set->exp;
    raw_scope = get_NBR(scope);
    raw_offset = get_NBR(offset);
    Main_Print(Leftpar_print_text);
    print_symbol(Pool_Set_P);
    Main_Print(Space_print_text);
    print_variable(raw_scope,
                   raw_offset);
    Main_Print(Space_print_text);
    print_expression(expression);
    Main_Print(Rightpar_print_text); }

static VOI_type print_set_local(STL_type Set)
  { EXP_type expression;
    NBR_type offset;
    UNS_type raw_offset;
    offset   = Set->ofs;
    expression = Set->exp;
    raw_offset = get_NBR(offset);
    Main_Print(Leftpar_print_text);
    print_symbol(Pool_Set_P);
    Main_Print(Space_print_text);
    print_variable(Scope,
                   raw_offset);
    Main_Print(Space_print_text);
    print_expression(expression);
    Main_Print(Rightpar_print_text); }

static VOI_type print_string(STR_type String)
  { TXT_type text;
    text = String->txt;
    FORMAT(String_print_text,
           text); }

static VOI_type print_symbol(SYM_type Symbol)
  { TXT_type text;
    text = Symbol->txt;
    FORMAT(Symbol_print_text,
           text); }

static VOI_type print_true(VOI_type)
  { Main_Print(True_print_text); }

static VOI_type print_unspecified(VOI_type)
  { Main_Print(Unspecified_print_text); }

static VOI_type print_vector(VEC_type Vector)
  { EXP_type value;
    UNS_type index,
             size;
    size = size_VEC(Vector);
    Main_Print(Leftvec_print_text);
    for (index = 1;
         index <= size;
         index++)
      { value = Vector[index];
        print_expression(value);
        if (index < size)
          Main_Print(Space_print_text); }
    Main_Print(Rightvec_print_text); }

static VOI_type print_while(WHI_type While)
  { EXP_type body,
             predicate;
    predicate  = While->prd;
    body       = While->bod;
    Main_Print(Leftpar_print_text);
    print_symbol(Pool_While_P);
    Main_Print(Space_print_text);
    Scope++;
    print_expression(predicate);
    indent();
    print_sequence(body);
    Scope--;
    Main_Print(Rightpar_print_text);
    outdent(); }

/*----------------------------------------------------------------------------*/

static VOI_type print_expression(EXP_type Expression)
  { TAG_type tag;
    tag = Grammar_Tag(Expression);
    switch (tag)
      { case APL_tag:
          print_application(Expression);
          return;
        case BEG_tag:
          print_begin(Expression);
          return;
        case CHA_tag:
          print_character(Expression);
          return;
        case DFF_tag:
          print_define_function(Expression);
          return;
        case DFV_tag:
          print_define_variable(Expression);
          return;
        case DFZ_tag:
          print_define_function_vararg(Expression);
          return;
        case FLS_tag:
          print_false();
          return;
        case GLB_tag:
          print_global(Expression);
          return;
        case IFF_tag:
          print_if_double(Expression);
          return;
        case IFZ_tag:
          print_if_single(Expression);
          return;
        case LCL_tag:
          print_local(Expression);
          return;
        case LMB_tag:
          print_lambda(Expression);
          return;
        case LMZ_tag:
          print_lambda_vararg(Expression);
          return;
        case NAT_tag:
          print_native(Expression);
          return;
        case NBR_tag:
          print_number(Expression);
          return;
        case NUL_tag:
          print_null();
          return;
        case PAI_tag:
          print_pair(Expression);
          return;
        case PRC_tag:
          print_procedure(Expression);
          return;
        case PRZ_tag:
          print_procedure_vararg(Expression);
          return;
        case QUO_tag:
          print_quote(Expression);
          return;
        case REA_tag:
          print_real(Expression);
          return;
        case STG_tag:
          print_set_global(Expression);
          return;
        case STL_tag:
          print_set_local(Expression);
          return;
        case STR_tag:
          print_string(Expression);
          return;
        case SYM_tag:
          print_symbol(Expression);
          return;
        case TRU_tag:
          print_true();
          return;
        case USP_tag:
          print_unspecified();
          return;
        case VEC_tag:
          print_vector(Expression);
          return;
        case WHI_tag:
          print_while(Expression);
          return;
        default:
          print_error_tag(IXT_message,
                          tag); }}

/*----------------------------- public functions -----------------------------*/

VOI_type Print_Display(EXP_type Value)
  { display_expression(Value); }

VOI_type Print_Initialize(VOI_type)
  { return; }

VOI_type Print_Newline(VOI_type)
  { display_newline(); }

VOI_type Print_Print(EXP_type Value)
  { Indentation = 0;
    Scope = 1;
    print_expression(Value); }