                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*   version 9: garbage collection   */
                     /*-----------------------------------*/
                     /*              reader               */
                     /*-----------------------------------*/

#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "SlipGrammar.h"

#include "SlipPool.h"
#include "SlipScan.h"
#include "SlipStack.h"

#include "SlipRead.h"

/*----------------------------- private constants ----------------------------*/

static const TXT_type IRB_message = "illegal right bracket";
static const TXT_type IRP_message = "illegal right parenthesis";
static const TXT_type ITK_message = "invalid token";
static const TXT_type IUP_message = "illegal use of period";
static const TXT_type MRP_message = "missing right parenthesis";
static const TXT_type PEP_message = "premature end of program";
static const TXT_type XCT_message = "excess tokens";

/*----------------------------- private variables ----------------------------*/

static SCA_type Token;

/*---------------------------- private prototypes ----------------------------*/

static EXP_type read_expression_C(VOI_type);

/*----------------------------- private functions ----------------------------*/

static EXP_type error(TXT_type Message)
  { Main_Error(Message);
    return Grammar_Null; }

static VOI_type next_token(VOI_type)
  { Token = Scan(); }

static EXP_type next_token_and_return(EXP_type Exp)
  { next_token();
    return Exp; }

/*----------------------------------------------------------------------------*/

static EXP_type read_bracket(VOI_type)
  { return error(IRB_message); }

static CHA_type read_character_C(VOI_type)
  { CHA_type character;
    CHR_type raw_character;
    Grammar_Default_Claim_C();
    raw_character = Scan_String[0];
    character = make_CHA_M(raw_character);
    return next_token_and_return(character); }

static EXP_type read_end_program(VOI_type)
  { return error(PEP_message); }

static FLS_type read_false(VOI_type)
  { return next_token_and_return(Grammar_False); }

static PAI_type read_list_C(VOI_type)
  { EXP_type expression;
    PAI_type pair;
    UNS_type count;
    next_token();
    for (count = 0;
         ;
         count++)
      if (Token == RPR_token)
        { Stack_Push_C(Grammar_Null);
          break; }
      else
      { expression = read_expression_C();
        Stack_Push_C(expression);
        if (Token == PER_token)
          { next_token();
            expression = read_expression_C();
            Stack_Push_C(expression);
            if (Token != RPR_token)
              return error(MRP_message);
            count++;
            break; }}
    next_token();
    for (pair = Stack_Pop();
         count > 0;
         count--)
      { SURVIVE_DEFAULT_C(pair);
        expression = Stack_Pop();
        pair = make_PAI_M(expression,
                          pair); }
    return pair; }

static NBR_type read_number(VOI_type)
  { DBL_type raw_number;
    NBR_type number;
    raw_number = strtoll(Scan_String, 0, 10);
    if (errno == ERANGE)
      return error("number out of range");
    number = Grammar_Number(raw_number);
		return next_token_and_return(number); }

static EXP_type read_parenthesis(VOI_type)
  { return error(IRP_message); }

static REA_type read_real_C(VOI_type)
  { FLO_type raw_real;
    REA_type real;
    raw_real = atof(Scan_String);
    Grammar_Default_Claim_C();
    real = make_REA_M(raw_real);
    return next_token_and_return(real); }

static EXP_type read_period(VOI_type)
  { return error(IUP_message); }

static EXP_type read_quote_C(VOI_type)
  { EXP_type expression;
    PAI_type pair;
    next_token();
		expression = read_expression_C();
    SURVIVE_DEFAULT_C(expression);
    pair = make_PAI_M(expression,
                      Grammar_Null);
    pair = make_PAI_M(Pool_Quote_P,
                      pair);
    return pair; }

static STR_type read_string_C(VOI_type)
  { STR_type string;
    UNS_type size;
    size = size_STR(Scan_String);
    Grammar_Dynamic_Claim_C(size);
    string = make_STR_M(Scan_String);
    return next_token_and_return(string); }

static SYM_type read_symbol_C(VOI_type)
  { SYM_type symbol;
    symbol = Pool_Enter_C(Scan_String);
    return next_token_and_return(symbol); }

static TRU_type read_true(VOI_type)
  { return next_token_and_return(Grammar_True); }

static VEC_type read_vector_C(VOI_type)
  { EXP_type expression;
    UNS_type count;
    VEC_type vector;
    next_token();
    if (Token == RBR_token)
      vector = Grammar_Empty_Vector_P;
    else
      { for (count = 0;
             Token != RBR_token;
             count++)
          { expression = read_expression_C();
 	          Stack_Push_C(expression); }
        Grammar_Dynamic_Claim_C(count);
        for (vector = make_VEC_M(count);
             count > 0;
             count--)
          vector[count] = Stack_Pop(); }
    return next_token_and_return(vector); }

static EXP_type read_expression_C(VOI_type)
  { switch (Token)
      { case CHA_token:
          return read_character_C();
        case END_token:
          return read_end_program();
        case FLS_token:
          return read_false();
        case IDT_token:
          return read_symbol_C();
        case LBR_token:
          return read_vector_C();
        case LPR_token:
          return read_list_C();
        case NBR_token:
          return read_number();
        case PER_token:
          return read_period();
        case QUO_token:
          return read_quote_C();
        case RBR_token:
          return read_bracket();
        case RPR_token:
          return read_parenthesis();
        case REA_token:
          return read_real_C();
        case STR_token:
          return read_string_C();
        case TRU_token:
          return read_true();
	      default:
          return error(ITK_message); }}

/*----------------------------- public functions -----------------------------*/

EXP_type Read_C(TXT_type Input)
  { EXP_type expression;
    Scan_Preset(Input);
    next_token();
    expression = read_expression_C();
    if (Token != END_token)
      return error(XCT_message);
    return expression; }

/*------------------------------ initialization ------------------------------*/

VOI_type Read_Initialize(VOI_type)
  { Scan_Initialize(); }
